#! /usr/bin/env python3
#
# google_word2vec.py: Create "deep learning" model for word vector representation
# using Google's word2vec algorithm (Mikolov et al. 2013) as implemented in Gensim.
# This can be used to support term similarity calculations.
#
# This also includes support for displaying term similarity based on the word2vec
# model for a list of phrases in the input. The similarity will be derived using
# all the terms together as positive cases as well as on just using each term
# individually.
#
# Notes:
# - (Mikolov et al. 2013): Mikolov, T., K. Chen, G. Corrado, and J. Dean (2013), 
#   "Efficient Estimation of Word Representations in Vector Space", ICLR workshop.
# - Based on http://radimrehurek.com/2014/02/word2vec-tutorial.
# - See gensim_test.py for script that supports document similarity instead of term similarity.
# - For reproducibility of results, the number of workers should be set to 1. Setting,
#   the random seed doesn't have any effect because gensim uses 1 by default.
#
# TODO:
# - *** Finalize the model to reduce memory footprint:
#   ex: model.init_sims(replace=True)
# - Have option to precompute top-n similarities for common words and save as shelve-like data file (e.g., to cut down on memory usage).
#
# TODO2:
# - Check gensim 4.0 API usage: https://github.com/piskvorky/gensim/wiki/Migrating-from-Gensim-3.x-to-4.
#
#------------------------------------------------------------------------
# Copyright (C) 2012-2018 Thomas P. O'Hara
#

"""Simple interface into Gensim's Word2vec algorithm"""

import argparse
import numpy
import os
import re
import sys
import logging
import multiprocessing

from gensim.models import Word2Vec
from mezcla import file_utils
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla import tpo_common as tpo
## OLD:
## import tpo_common as tpo
## import glue_helpers as gh

WORD2VEC_MODEL_EXT = ".word2vec"
NUM_TOP = tpo.getenv_integer("NUM_TOP", 5, "Maximum number of related terms to display")
SKIP_TERMS_WITH_PUNCTUATION = tpo.getenv_boolean("SKIP_TERMS_WITH_PUNCTUATION", False, "Omit related terms with punctuation characters")
# TODO: rework so that SKIP_LOW_FREQUENCY_TERMS conditional upon TERM_FREQ_FILE
SKIP_LOW_FREQUENCY_TERMS = tpo.getenv_boolean("SKIP_LOW_FREQUENCY_TERMS", False, "Omit related terms that have low frequency")
DEFAULT_TERM_FREQ_FILE = "term.freq" if SKIP_LOW_FREQUENCY_TERMS else ""
TERM_FREQ_FILE = tpo.getenv_text("TERM_FREQ_FILE", DEFAULT_TERM_FREQ_FILE, "Frequency for term occurrences")
TERM_FREQ_HASH = tpo.create_lookup_table(TERM_FREQ_FILE) if gh.non_empty_file(TERM_FREQ_FILE) else {}
NUM_WORKERS = tpo.getenv_integer("NUM_WORKERS", multiprocessing.cpu_count(), "Number of worker threads; use 1 to reproduce results")
RANDOM_SEED = tpo.getenv_integer("RANDOM_SEED", -1, 
                                 "Integral seed for random number generation")
PRESERVE = tpo.getenv_boolean("PRESERVE", False,
                              "Preserve format of text")
DOWNCASE = tpo.getenv_boolean("DOWNCASE", not PRESERVE,
                              "Convert text to lowercase")
SKIP_INDIVIDUAL = tpo.getenv_boolean("SKIP_INDIVIDUAL", False,
                                     "Omit similarity for individual tokens in input sentences")
## TODO:
## JSON_FORMAT = system.getenv_bool(
##     "JSON_FORMAT", False,
##     desc="USe JSON format for output such as similarity")
SIM_OUTPUT_FILE = system.getenv_value(
    "SIM_OUTPUT_FILE", None,
    desc="Output file for similarity info for all terms")

def format_related_terms(model, positive_terms, max_num=NUM_TOP):
    """Determine related terms from MODEL for POSITIVE_TERMS, returning at most MAX_NUM entries each."""
    # Try to get most similar terms. If words are not in the vocabulary
    # try with the remainder if any.
    all_related_info = []
    try:
        ## OLD: all_related_info = model.most_similar(positive=positive_terms)
        all_related_info = model.wv.most_similar(positive=positive_terms)
    except KeyError:
        missing = [w for w in positive_terms if w not in model]
        tpo.print_stderr("Warning: omitting words not in model: %s" % missing)
        ok_words = tpo.difference(positive_terms, missing)
        if ok_words:
            try:
                all_related_info = model.wv.most_similar(positive=ok_words)
            except:
                tpo.print_stderr("Unexpected error in format_related_terms: " + str(sys.exc_info()))

    # Add related terms unless filtered due to low frequency or embedded punctuation
    related_specs = []
    for (term, score) in all_related_info:
        if SKIP_LOW_FREQUENCY_TERMS and term.lower() not in TERM_FREQ_HASH:
            tpo.debug_print("Skipping low frequency related term '%s'" % term, 6)
            continue
        if SKIP_TERMS_WITH_PUNCTUATION and re.search(r"\W", term):
            tpo.debug_print("Skipping related term '%s' due to punctuation" % term, 6)
            continue
        related_specs.append(term + ": " + tpo.round_num(score))
        if len(related_specs) == max_num:
            break
    return ", ".join(related_specs)

def tokenize(text):
    r"""Tokenize TEXT according to regex word tokens (i.e., \W+), which defaults to [A-Za-z0-9_]+"""
    # TODO: Allow for tokenization regex to be overwritten
    token_regex = r"(\W+)" if not PRESERVE else r"(\S+)"
    tokens = [t.strip() for t in re.split(token_regex, text) if t.strip()]
    if DOWNCASE:
        tokens = [t.lower() for t in tokens]
        
    tpo.debug_format("tokenize({txt!r}) => {t!r}", 7, txt=text, t=tokens)
    return tokens


class MySentences(object):
    """Class for processing files line by line. Note: the input file should have one document per line, and the text will remain case sensitive in word2vec."""
    # Note: based on Sentences class from gensim sample.
    # TODO: Rename to Documents for clarity; make downcase an option of the class

    def __init__(self, file_name):
        """Class constructor: FILE_NAME is text file or directory"""
        tpo.debug_format("MySentences.__init__({f})", 6, f=file_name)
        self.file_name = file_name
        return

    def __iter__(self):
        """Returns iterator producing one line at a time"""
        # Derive the list of filenames to process
        # TODO: support recursive directory descent
        tpo.debug_print("in MySentences.__iter__()", 6)
        file_names = None
        if os.path.isdir(self.file_name):
            dir_name = self.file_name
            file_names = [os.path.join(dir_name, f) for f in os.listdir(dir_name)]
        else:
            file_names = [self.file_name]

        # Feed each sentence individually from each file
        # TODO: add preprocessing (e.g., tokenize, make lowercase, etc.)
        for file_name in file_names:
            if os.path.isdir(file_name):
                tpo.debug_format("Warning: skipping subdirectory {f}", tpo.WARNING, f=file_name)
                continue
            tpo.debug_format("Processing file {f}", tpo.DETAILED, f=file_name)
            for line in system.open_file(file_name):
                ## OLD: tokens = line.split()
                tokens = tokenize(line)
                tpo.debug_format("MySentences.__iter__: yielding {t}", 6, t=tokens)
                yield tokens
        tpo.debug_print("out MySentences.__iter__()", 6)
        return


def main():
    """Entry point for script"""
    tpo.debug_print("main(): sys.argv=%s" % sys.argv, 4)

    # Parse command-line arguments
    env_options = tpo.formatted_environment_option_descriptions(indent="  ")
    usage_description = tpo.format("""
Creates Google word2vec model (via gensim) of word distributions inferrred from 
the occurrences in the input text file. Note: input should be a text file 
(or directory) when creating from scratch or the basename of model file 
if loading existing model.

Notes:
- The input file should have one document per line (multiple sentences allowed).
- The following environment options are available:
  {env}
    """, env=env_options)
    parser = argparse.ArgumentParser(description=usage_description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--save", default=False, action='store_true', help="Save model to disk")
    parser.add_argument("--load", default=False, action='store_true', help="Load model from disk")
    parser.add_argument("--list-terms", default=False, action='store_true', help="Print vocabulary")
    parser.add_argument("--print", default=False, action='store_true', help="Print vectors on standard output")
    parser.add_argument("filename", default=None, help="Input data filename (or basename when loading previously saved model); if a directory all files within are processed")
    parser.add_argument("--output-basename", default=None, help="Basename to use for output (by default input file without .txt extension)")
    parser.add_argument("--show-similarity", default=False, action='store_true', help="Show similar terms for those from input (one per line)")
    # TODO: parser.add_argument("--language-model", default=None, help="Language model to use for rating similar terms")
    args = vars(parser.parse_args())
    tpo.debug_print("args = %s" % args, 5)
    filename = args['filename']
    save = args['save']
    load = args['load']
    list_terms = args['list_terms']
    print_vectors = args['print']
    show_similarity = args['show_similarity']
    output_basename = args['output_basename']
    # TODO: put version of glue_helper's assertion into tpo_common.py already!
    gh.assertion(filename)

    # Derive the basename if not given (checking one of .txt/.list/.prep extensions if training or .word2vec if loading)
    # TODO: rework in terms of stripping whatever file extension is used (e.g., "it.fubar" => "it")
    if not output_basename:
        input_extensions = [".txt", ".list", ".prep"] if (not load) else [WORD2VEC_MODEL_EXT]
        output_basename = filename
        for extension in input_extensions:
            output_basename = gh.remove_extension(filename, extension)
            if (output_basename != filename):
                break
    tpo.debug_print("output_basename=%s" % output_basename, 5)

    # Enable logging if debugging
    if (tpo.debugging_level()):
        # TODO: use mapping from symbolic LEVEL user option (e.g., via getenv)
        level = logging.INFO if (tpo.debug_level < 4) else logging.DEBUG
        logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=level)

    # Optionally set random seed
    if RANDOM_SEED != -1:
        tpo.debug_format("Setting random seed to {RANDOM_SEED}")
        numpy.random.seed(RANDOM_SEED)

    # Process the input file(s), either creating model from scratch or loading existing one
    if load:
        model = Word2Vec.load(filename)
    else:
        sentences = MySentences(filename)
        if tpo.verbose_debugging():
            # TODO: try to develop develop read-only function that makes copy of iterator
            sentences = list(sentences)
            gh.assertion(len(sentences) > 0)
            tpo.debug_format("sentences={s}", 6, s=sentences)
        # Notes: 1 is default for word2vec (todo, try None)
        seed = 1 if (RANDOM_SEED == -1) else RANDOM_SEED
        model = Word2Vec(sentences, workers=NUM_WORKERS, seed=seed)

        # Optionally save model to disk
        if (save):
            model.save(output_basename + WORD2VEC_MODEL_EXT)

    # Print the vocabulary
    if list_terms:
        all_words = sorted(model.wv.key_to_index.keys())
        print("\n".join(all_words))
            
    # Print the vector representations
    # TODO: add option to print word similarity matrix
    if print_vectors:
        ## OLD: all_words = sorted(model.vocab.keys())
        all_words = sorted(model.wv.key_to_index.keys())
        tpo.debug_format("model={m}", 6, m=model)
        print("Vocaulary terms: %s" % all_words)
        for word in all_words:
            ## OLD:
            ## tpo.debug_format("model[%s]=%s" % (word, model[word]), 5)
            ## print("%s\t%s" % (word, model[word]))
            tpo.debug_format("model[%s]=%s" % (word, model.wv[word]), 5)
            print("%s\t%s" % (word, model.wv[word]))

    # Show similarity info for terms from input
    # TODO: add better recovery for terms unknown
    if show_similarity:
        tpo.debug_print("Show similarity for terms from stdin", 4)
        print("term(s): similarity info")
        for line in sys.stdin:
            ## OLD: terms = [t.strip() for t in re.split(r"\W+", line.strip().lower())]
            terms = tokenize(line)
            try:
                # TODO: shows language model score for terms replaced by related terms
                if not terms:
                    pass
                elif len(terms) > 1 or SKIP_INDIVIDUAL:
                    print("[%s]: %s" % (", ".join(terms), format_related_terms(model, terms)))
                else:
                    if not SKIP_INDIVIDUAL:
                        for term in terms:
                            print("[%s]: %s" % (term, format_related_terms(model, [term])))
                print("")
            except KeyError:
                tpo.print_stderr("Error: %s" % str(sys.exc_info()))
    if SIM_OUTPUT_FILE:
        sim_info = {}
        for term in sorted(model.wv.key_to_index.keys()):
            sim_info[term] = model.wv.most_similar(positive=[term])
        file_utils.write_json(SIM_OUTPUT_FILE, sim_info, default=str)
    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# coding: utf-8
# TODO:
# - Fix env-spec above and put coding on same line.
# - Verify the fixup for PEP-0479 didn't break anything in Python 2 (e.g., for regression testing).
# - *** Remove Python2-specific code!
#

"""Pre-processing step for text.

Example:
    >>> import mezcla.tfidf.preprocess as mtp
    >>> p = mtp.Preprocessor(gramsize=2)
    >>> sorted(list([k.text for k in p.yield_keywords('a b c')]))
    ['a b', 'b c']
"""
# Note: uses import ... as for sake of reload
# TPO: TODO: 
# - add logging
# - replace all_ngrams with min_ngram_size
# - add test set and check explicitly with multi-threaded access
from __future__ import absolute_import, with_statement

# Standard modules
from collections import namedtuple
import re
import sys  # python2 support
from threading import RLock
import html

# Installed modules
## TODO: find better documented memoization package
from cachetools import LRUCache, cached  # python2 support
import nltk
from nltk.stem import SnowballStemmer
## OLD: from sklearn.feature_extraction.text import CountVectorizer
CountVectorizer = None
from stop_words import get_stop_words

# Local modules
from mezcla import debug
from mezcla.my_regex import my_re, REGEX_TRACE_LEVEL
from mezcla import system
from mezcla.tfidf.config import BASE_DEBUG_LEVEL as BDL
from mezcla.tfidf.dockeyword import DocKeyword

unescape = html.unescape


# TPO: Allow word splitting pattern to be overwritten.
# TODO: Get WORD_REGEX from environment.
# TODO: Have option to ignore words with specific punctuation (or to include).
# TODO: strip period to allow for ngrams with abbreviations (e.g., "Dr. Jones").
SPLIT_WORDS = system.getenv_bool(
    "SPLIT_WORDS", False,
    description="Split by word token instead of whitespace")
TFIDF_ALLOW_PUNCT = system.getenv_bool(
    "TFIDF_ALLOW_PUNCT", False,
    description="Allow ngrams to include words with punctuation")
TFIDF_SENT_SPLITTER = system.getenv_bool(
    "TFIDF_SENT_SPLITTER", False,
    description="Use NLTK for splitting sentences for TF/IDF")
USE_SIMPLE_SENT_SPLITTER = (not TFIDF_SENT_SPLITTER)
BAD_WORD_PUNCT_REGEX = system.getenv_text(
    "BAD_WORD_PUNCT_REGEX", r"[^a-zA-Z0-9_'$ -]",
    description="Regex for punctuation not in words")
SKIP_WORD_CLEANING = system.getenv_bool(
    "SKIP_WORD_CLEANING", False,
    description="Omit word punctuation cleanup")
USE_SKLEARN_COUNTER = system.getenv_bool(
    "USE_SKLEARN_COUNTER", False,
    description="Use sklearn CountVectorizer for ngrams")
TFIDF_PRESERVE_CASE = system.getenv_bool(
    "TFIDF_PRESERVE_CASE", False,
    description="Preserve case in TFIDF ngrams")
TFIDF_LANGUAGE = system.getenv_value(
    "TFIDF_LANGUAGE", None,
    description="Language for text preprocessing")

if SPLIT_WORDS:
    debug.trace(2, "FYI: Splitting by word token (not whitespace)\n")
    debug.assertion(not TFIDF_ALLOW_PUNCT)
WORD_REGEX = r'\w+' if SPLIT_WORDS else r'\S+'


def handle_unicode(text):
    """Needed for the description fields."""
    # EX: handle_unicode("Predactica\u2122") => "Predactica \u2122"
    if re.search(r'\\+((u([0-9]|[a-z]|[A-Z]){4}))', text):
        text = text.encode('utf-8', 'ignore').decode('unicode-escape', 'ignore')
    text = re.sub(r'\\n', '\n', text)
    text = re.sub(r'\\t', '\t', text)
    # Put space around non-ascii unicode characters
    # note: excludes special punctuation like U+00A9 (©)
    EXT_ASC_CH = "\u0000-\u007F\u00C0-\u0100"
    if re.search(fr'[^{EXT_ASC_CH}]', text):
        text = re.sub(fr'([{EXT_ASC_CH}]) ([^{EXT_ASC_CH}]) ([{EXT_ASC_CH}])',
                      r'\1 \2 \3', f" {text} ", flags=re.VERBOSE)
        text = text.strip(" ")
    return text


def handle_html_unquote(text):
    """Detect if there are HTML encoded characters, then decode them."""
    if re.search(r'(&#?x?)([A-Z]|[a-z]|[0-9]){2,10};', text):
        text = unescape(text)
    return text


def handle_mac_quotes(text):
    """Handle the unfortunate non-ascii quotes OSX inserts."""
    text = text.replace('“', '"').replace('”', '"')\
        .replace('‘', "'").replace('’', "'")
    return text


def handle_text_break_dash(text):
    """Convert text break dashes into semicolons to simplify things.

    For example,
        "She loved icecream- mint chip especially"
        "She loved icecream - mint chip especially"
    will both convert to
        "She loved icecream; mint chip especially".

    However,
        "The 27-year-old could eat icecream any day"
    will not be changed.
    """
    # EX: handle_text_break_dash("a - b") => "a; b"
    return re.sub(r'\s+-\s*|\s*-\s+', '; ', text)


def clean_text(raw_text):
    """Strip text of non useful characters.
    Note: This is no-op if SKIP_WORD_CLEANING (global).
    """
    # TODO: rework so SKIP_WORD_CLEANING applied elsewhere
    text = raw_text

    
    if not SKIP_WORD_CLEANING:
        # Must strip HTML tags out first!
        text = re.sub('<[^<]+?>', ' ', text)
        text = handle_unicode(text)
        text = handle_html_unquote(text)
        text = handle_mac_quotes(text)
        text = handle_text_break_dash(text)
        if not TFIDF_PRESERVE_CASE:
            text = text.lower()

        ## TODO3: make &-stripping optional
        regex_subs = ['\t', '\n', '\r', r'\s+', '&']
        for regex_sub in regex_subs:
            text = re.sub(regex_sub, ' ', text)
    debug.trace(BDL + 2, f"clean_text({raw_text!r}) => {text!r}")
    return text


def create_stemmer(language):
    """Return wordform stemmer for LANGUAGE (via SnowBall)"""
    stemmer = SnowballStemmer(language)
    
    # Define the stemmer for the particular language
    def stem_wordform(wordform):
        """Returned root of WORDFORM"""
        root = stemmer.stem(wordform)
        debug.trace_fmt(BDL + 3, "stem_wordform({wf}) => {r}", wf=wordform, r=root)
        return root

    # Do sanity check
    if (language == "english"):
        debug.assertion(stem_wordform("running") == "run")

    return stem_wordform


class Preprocessor(object):
    """Prep the text for TF-IDF calculations.

    Fixes some unicode problems, handles HTML character encoding,
    and removes HTML tags.

    Strips some non alphanumeric characters, but keeps ngram boundary
    markers (eg, period (',') and semi-colon (';'))

    If a stopwords file is provided, it will remove stopwords.

    Example:
        >>> import mezcla.tfidf.preprocess as mtp
        >>> processor = mtp.Preprocessor(language='english')
        >>> processor.handle_stopwords('She is an interesting gal.', clean=True)
        'interesting gal.'
        >>> sp_processor = mtp.Preprocessor(language='spanish')
        >>> sp_processor.handle_stopwords('She is an interesting gal.', clean=True)
        'she is an interesting gal.'
    """

    stopwords = set()
    contractions = r"(n't|'s|'re|'ll)$"
    ## OLD: negative_gram_breaks = r'[^:;!^,\?\.\[|\]\(|\)"`]+'
    negative_gram_breaks = r'[^:;!^,\?\.\[|\]\(|\)"`]+' if not TFIDF_ALLOW_PUNCT else  r'[^:;!\?\.]+'
    supported_languages = (
        'danish', 'dutch', 'english', 'finnish', 'french', 'german', 'hungarian',
        'italian', 'kazakh', 'norwegian', 'porter', 'portuguese', 'romanian',
        'russian', 'spanish', 'swedish', 'turkish'
    )

    def __init__(self, min_ngram_size=None, max_ngram_size=None,
                 language=None, stopwords_file=None, stemmer=None,
                 gramsize=None, all_ngrams=None, use_sklearn_counter=None):
        """Preprocessor must be initalized for use if using stopwords.

        stopwords_file (filename): contains stopwords, one per line
        stemmer (function):  takes in a word and returns the stemmed version
        min_ngram_size (int-or-None):
            if integral, gives the lower bound
        max_ngram_size (int-or-None):
            if integral, gives the upper bound
        negative_gram_breaks (regex):
            if a word ends with one of these characters, an
            ngram may not cross that. Expressed as a _negative_ regex.
            Example:
            in the sentence "Although he saw the car, he ran across the street"
            "car he" may not be a bi-gram
        stopwords_file (filename):
            Provide a list of stopwords. If used in addition to "language", the
            provided stopwords file overrides the default.
        language:
            name of language from NLTK (n.b., use empty string or 'n/a' for none)
        stemmer (function):
            A function that takes in a single argument (str) and returns a string
            as the stemmed word. Overrides the default behavior if specified.
            Default None:
                Use the NLTK snowball stemmer for the sepcified language. If
                language is not found, no stemming will take place.
        gramsize (int-or-None): maximum word size for ngrams
            Note: deprecated (use min_ngram_size instead).
        all_ngrams (bool):
            if true, all possible ngrams of length "gramsize" and smaller will
            be examined. If false, only ngrams of _exactly_ length "gramsize"
            will be run.
            Note: deprecated (use min_ngram_size instead).
        """
        # Note: stemmer is no-op if no language specified
        self.__stemmer = None
        # Note: language can be empty string to block defaults
        if language is None:
            language = TFIDF_LANGUAGE
        if (language and (language != "n/a")):
            debug.assertion(language in self.supported_languages)
            if language in SnowballStemmer.languages:
                self.__stemmer = create_stemmer(language)
            self.stopwords = get_stop_words(language)
        if stopwords_file:
            self._load_stopwords(stopwords_file)
        if stemmer:
            self.__stemmer = stemmer
        if not self.__stemmer:
            debug.trace(BDL - 2, "Warning: defining no-op stemmer in Preprocessor")
            self.__stemmer = lambda x: x  # no change to word
        if use_sklearn_counter is None:
            use_sklearn_counter = USE_SKLEARN_COUNTER
        self.use_sklearn_counter = USE_SKLEARN_COUNTER
        debug.assertion(not (gramsize and max_ngram_size))
        debug.assertion(not (all_ngrams and min_ngram_size))
        self.__gramsize = (max_ngram_size or gramsize or 1)
        self.__all_ngrams = all_ngrams
        self.__min_ngram_size = (min_ngram_size or gramsize or 1)

    @property
    def gramsize(self):
        """Number of words in the ngram."""
        return self.__gramsize

    @property
    def all_ngrams(self):
        """True if ngrams of size "gramsize" or smaller will be generated.

        False if only ngrams of _exactly_ size "gramsize" are generated.
        """
        return self.__all_ngrams

    @property
    def min_ngram_size(self):
        """If non-Null, can be used to override all_ngrams (q.v.), so
        ngrams go from [M, N] rather than [1, N]
        """
        return self.__min_ngram_size

    def _load_stopwords(self, filename):
        with system.open_file(filename) as f:
            words = []
            for line in f:
                words.append(line.strip())
        self.stopwords = set(words)

    def handle_stopwords(self, text, clean=None):
        """Remove stop words from the text."""
        out = []
        if clean:
            text = clean_text(text)
        ## TODO2: lower().split()
        for word in text.split(' '):
            # Remove common contractions for stopwords when checking list
            check_me = re.sub(self.contractions, '', word)
            if check_me in self.stopwords:
                continue
            out.append(word)
        return ' '.join(out)

    def normalize_term(self, text):
        """Cleans the text characters (optional), removes stopwords, and applies stemming

        Assumes the input is already the number of words you want for the ngram.
        Note: Omits word punctuation cleanup if SKIP_WORD_CLEANING (global).
        """
        text = clean_text(text)
        text = self.handle_stopwords(text)
        return self.stem_term(text)

    ## TODO: do performance tests to select suitable cache size
    @cached(LRUCache(maxsize=100000), lock=RLock())
    def _stem(self, word):
        """The stem cache is used to cache up to 100,000 stemmed words.

        This substantially speeds up the word stemming on larger documents.
        """
        return self.__stemmer(word)

    def stem_term(self, term):
        """Apply the standard word procesing (eg stemming). Returns a stemmed ngram."""
        return ' '.join([self._stem(x) for x in term.split(' ')])

    def yield_keywords(self, raw_text, document=None):
        """Yield keyword objects as mono, di, tri... *-grams.

        Use this as an iterator.

        Will not create ngrams across logical sentence breaks.
        Example:
            s = "Although he saw the car, he ran across the street"
            the valid bigrams for the sentences are:
            ['Although he', 'saw the', 'he saw', 'the car',
            'he ran', 'across the', 'ran across', 'the street']
            "car he" is not a valid bi-gram

        This will also stem words when applicable.
        Example:
            s = "All the cars were honking their horns."
            ['all', 'the', 'car', 'were', 'honk', 'their', 'horn']
        """
        ## TODO: yield_method = self.slow_yield_keywords if not USE_SKLEARN_COUNTER else self.yield_sklearn_keywords
        if self.use_sklearn_counter:
            result = self.quick_yield_keywords(raw_text, document=document)
        else:
            result = self.full_yield_keywords(raw_text, document=document)
        return result

    def re_search(self, regex, text):
        """Invoke my_re.search with higher tracing"""
        debug.reference_var(self)
        TRACE_LEVEL = (REGEX_TRACE_LEVEL + 1)
        return my_re.search(regex, text, base_trace_level=TRACE_LEVEL)
    
    def full_yield_keywords(self, raw_text, document=None):
        """Full-featured version of keyword generation, including support for offsets"""
        if sys.version_info[0] < 3:  # python2 support
            if isinstance(raw_text, str):
                raw_text = raw_text.decode('utf-8', 'ignore')
        ## TPO: HACK: support min ngram size
        gramlist = None
        if self.all_ngrams:
            gramlist = range(1, self.gramsize + 1)
        if self.min_ngram_size:
            debug.assertion(not gramlist)
            if gramlist:
                sys.stderr.write("Error: ignoring obsolete all_ngrams and using min_ngram_size\n")
            gramlist = range(self.min_ngram_size, self.gramsize + 1)
        if not gramlist:
            gramlist = [self.gramsize]
        debug.trace_fmt(BDL + 2, "gramlist={gl}", gl=gramlist)

        sentence_split = (positional_splitter(self.negative_gram_breaks, raw_text)
                          if USE_SIMPLE_SENT_SPLITTER else nltk_sent_splitter(raw_text))
        for sentence in sentence_split:
            debug.trace_expr(BDL + 2, sentence.text)
            words = positional_splitter(WORD_REGEX, sentence.text)
            # Remove all stopwords
            words_no_stopwords = []
            for w in words:
                # Remove common contractions for stopwords when checking list
                check_me = re.sub(self.contractions, '', w.text)
                if check_me not in self.stopwords:
                    words_no_stopwords.append(w)
            if debug.debugging(BDL + 2):
                words_text_no_stopwords = [w.text for w in words_no_stopwords]
                debug.trace_expr(BDL + 2, words_text_no_stopwords)

            # Make the ngrams
            # TODO: make sure stripped stopwords block ngram (e.g. "dog and cat" =/=> "dog cat")
            for gramsize in gramlist:
                # You need to try as many offsets as chunk size
                # TODO: Allow stopwords in middle of ngram of size 3 or more
                for offset in range(0, gramsize):  # number of words offset
                    data = words_no_stopwords[offset:]
                    text_in_chunks = [data[pos:pos + gramsize]
                                      for pos in range(0, len(data), gramsize)
                                      if len(data[pos:pos + gramsize]) == gramsize]
                    for word_list in text_in_chunks:
                        word_text = ' '.join([self.stem_term(w.text) for w in word_list])
                        ## OLD: if (SPLIT_WORDS or (not re.search(BAD_WORD_PUNCT_REGEX, word_text))):
                        exclude = self.re_search(BAD_WORD_PUNCT_REGEX, word_text)
                        if TFIDF_ALLOW_PUNCT and exclude:
                            # Include unless punctuation starts or ends the text or is after a space
                            # TODO2: add BAD_NGRAM_PUNCT_REGEX (e.g., "[;!?]")
                            exclude = (self.re_search("^" + BAD_WORD_PUNCT_REGEX, word_text) or
                                       self.re_search(BAD_WORD_PUNCT_REGEX + "$", word_text) or
                                       self.re_search(" " + BAD_WORD_PUNCT_REGEX, word_text))
                        if not exclude:
                            word_global_start = sentence.start + word_list[0].start
                            word_global_end = sentence.start + word_list[-1].end
                            yield DocKeyword(word_text, document=document, start=word_global_start, end=word_global_end)
                        else:
                            debug.trace(BDL + 3, f"Ignoring {gramsize}-gram {word_text!r}")
        return

    def quick_yield_keywords(self, raw_text, document=None):
        """Quick version for yielding keywords, using sklearn for ngram generation
        Note: the DocKeyword objects don't include offset information"""
        # TODO1: support BAD_WORD_PUNCT_REGEX
        # Do dynamic load(s)
        global CountVectorizer
        if not CountVectorizer:
            # pylint: disable=import-outside-toplevel,redefined-outer-name
            from sklearn.feature_extraction.text import CountVectorizer
        vectorizer = CountVectorizer(ngram_range=(self.min_ngram_size, self.gramsize))
        analyzer = vectorizer.build_analyzer()
        ## DEBUG: debug.trace_expr(8, analyzer)
        for ngram in analyzer(raw_text):
            ## DEBUG: debug.trace_expr(9, ngram)
            yield DocKeyword(ngram, document=document)
        return


PositionalWord = namedtuple('PositionalWord', ['text', 'start', 'end'])

def positional_splitter(regex, text):
    r"""Yield sentence chunks (as defined by the regex) as well as their location.

    NOTE: the regex needs to be an "inverse match"
    Example:
        To split on whitespace, you match:
        r'\S+'  <-- "a chain of anything that's NOT whitespace"
    """
    ## TODO2: treat -- as clausal punctuation if TFIDF_ALLOW_PUNCT
    for res in re.finditer(regex, text):
        ## TODO: drop try/except
        try:
            yield PositionalWord(res.group(0), res.start(), res.end())
        except StopIteration:
            pass
    return


def nltk_sent_splitter(text):
    """Yield sentence chunks as well as their location."""
    # Note: Similar to positional_splitter but uses NLTK instead of negative_gram_breaks regex
    offset = 0
    for sent in nltk.tokenize.sent_tokenize(text):
        end = offset + (1 + len(sent))
        yield PositionalWord(sent, offset, end)
        offset = end
    return


def main():
    """Entry point for script: just runs a simple test"""
    text = "my man fran is not a man"
    p = Preprocessor(gramsize=2)
    debug.trace_object(BDL - 2, p)
    ngrams = list(p.text for p in p.yield_keywords(text))
    debug.trace_expr(3, text, ngrams)
    debug.assertion("my man" in ngrams)
    debug.assertion("my man fran" not in ngrams)
    debug.assertion("fran man" not in ngrams)
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone. A simple test will be run.")
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Mezcla is Spanish for mixture, and this repository contains a variety of Python scripts.

Miscellaneous Python scripts developed over the course of several independent consulting projects. This also includes some code samples I adapted from publicly available source. (The code is not proprietary in nature. For example, it was not "borrowed" from proprietary source files, nor based on proprietary processes.)

Spoiler alter: this is not "Pythonic python": I'm more into R&D than production programming. Nonetheless, there's a some useful scripts here, so I made the repository available. It is public in the spirit of open source software. 
 
This repository is licensed under the GNU Lesser General Public Version 3 (LGPLv3). See LICENSE.txt.

Adhoc usage:

    from mezcla import *
    debug.trace(TL.DEFAULT, "Hey")
    debug.trace(TL.DETAILED, "Joe")

Tom O'Hara
Feb 2022
"""
version = "1.4.0.4"
__VERSION__ = version
__version__ = __VERSION__

# Standard module(s)
import sys
## DEBUG: sys.stderr.write(f"{__file__=}\n")

# Note: requires python 3 or higher
PYTHON3_PLUS = (sys.version_info[0] >= 3)
assert PYTHON3_PLUS, "Python 3 or higher: ¡por favor!"

# Local modules
# Define most common imports for causual usage
# Note: avoids syntax error is debug and system import directly
## TODO
## debug = None
## mezcla = None
## system = None
## if PYTHON3_PLUS:
##     ## TODO: get relative imports to work
##     ## from . import debug
##     ## from .mezcla import debug
##     import mezcla
##     from mezcla import debug
##     from mezcla import system
##     TL = debug.TL
##     ## DEBUG: debug.trace_expr(TL.DEFAULT, debug, mezcla, system, TL)
## else:
##     TL = None
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Constants
TL = debug.TL

# Expose commonly used modules
__all__ = ["debug", "gh", "my_re", "system", "TL", "__VERSION__"]

## OLD:
## if __name__ == '__main__':
##     debug.trace(TL.USUAL, f"Version: {__VERSION__}")
##     system.print_error(f"Warning: {__file__} is not intended to be run standalone\n")
## NOTE: See https://stackoverflow.com/questions/43393764/python-3-6-project-structure-leads-to-runtimewarning
debug.trace(TL.DETAILED, f"mezcla version: {__VERSION__}")



#! /usr/bin/env python3
#
# Common code and settings used in tests.
#

"""Common test module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
## TODO: import pytest

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Constants and environment options
# Note: These are just intended for internal options, not for end users.
# It also allows for enabling options in one place.
#
USER = system.getenv_bool(
    "USER", "user",
    description="Current user")
HOME = system.getenv_bool(
    "HOME", "/home/user",
    description="Home directory")
UNDER_RUNNER = system.getenv_bool(
    "UNDER_RUNNER", HOME == "/home/runner",
    description="Whether running under Github actions")
RUN_SLOW_TESTS = system.getenv_bool(
    "RUN_SLOW_TESTS", 
    description="Alias for not[-]SKIP_SLOW_TESTS")
SKIP_SLOW_TESTS = system.getenv_bool(
    "SKIP_SLOW_TESTS",
    (not (UNDER_RUNNER or RUN_SLOW_TESTS)),
    description="Omit tests that can a while to run")
SKIP_SLOW_REASON="Ignoring slow test"
SKIP_UNIMPLEMENTED_TESTS = system.getenv_bool(
    # Note: used to avoid clutter due to (so many) unimplemented tests, such as
    # when using --runxfail for better test failure diagnostics.
    "SKIP_UNIMPLEMENTED_TESTS", 
    False,
    description="Skip tests not yet implemented")
SKIP_UNIMPLEMENTED_REASON = "Ignoring unimplemented test"
SKIP_EXPECTED_REASON = "Skipping cases that should never pass (e.g., intentional error)"
SKIP_EXPECTED_ERRORS = system.getenv_bool(
    # Note: this helps filter known errors before running error checking script,
    # (e.g., check_errors.py in companion repo tomasohara/shell-scripts).
    # It is different from xfail in that the tests are not likely to ever pass.
    "SKIP_EXPECTED_ERRORS",
    False,
    description="Skip cases intentionally causing conversion errors, etc.")

# Globals
mezcla_root_dir = None

#-------------------------------------------------------------------------------

def fix_indent(code):
    """Make sure CODE indented proper if it is a string;
    however, list input is returned as is.
    Note: this accounts for code defined with indented triple-quoted strings
    
    >>> fix_indent('''
                   print("ok")
                   ''')
    'print("ok")'
    """
    result = code
    if isinstance(result, str) and my_re.search(r"^\n( +)", result):
        indentation = my_re.group(1)
        result = my_re.sub(fr"^{indentation}", "", result, flags=my_re.MULTILINE)
    debug.trace(8, f"fix_indent{code!r} => {result!r}")
    return result


def get_mezcla_root_dir():
    """Get the base directory for the mezcla distribution"""
    test_dir = gh.dir_path(__file__)
    root_dir = gh.real_path(gh.form_path(test_dir, "..", ".."))
    debug.assertion(system.file_exists(gh.form_path(root_dir, "LICENSE.txt")))
    debug.trace(5, f"get_mezcla_root_dir() => {root_dir!r}")
    return root_dir      

mezcla_root_dir = get_mezcla_root_dir()

#-------------------------------------------------------------------------------

if __name__ == "__main__":
    debug.trace_current_context()
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone\n")



#! /usr/bin/env python3
#
# Utilities for working with GPU's such as via PyTorch or using NVIDIA commands.
#
#

"""GPU support"""

# Standard modules
from typing import Optional

# Installed modules
import torch

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system

# Constants
TL = debug.TL
DEBUG_LEVEL = debug.get_level()
#
# Environment-based
HAS_CUDA = torch.cuda.is_available()
HAS_MPS = (hasattr(torch.backends, "mps") and torch.backends.mps.is_available())
USE_CPU = system.getenv_bool(
    "USE_CPU", False,
    description="Uses Torch on CPU if true")
## OLD: TORCH_DEVICE_DEFAULT = ("cpu" if (USE_CPU or not HAS_CUDA) else "cuda")
TORCH_DEVICE_DEFAULT = ("cuda" if HAS_CUDA else "mps" if HAS_MPS else "cpu")
TORCH_DEVICE = system.getenv_text(
    "TORCH_DEVICE", TORCH_DEVICE_DEFAULT,
    description="Device for running torch"
)
GPU_DEBUG_LEVEL = system.getenv_int(
    "GPU_DEBUG_LEVEL", TL.DETAILED,
    desc="Default trace level for trace_gpu_usage, etc.")

#-------------------------------------------------------------------------------

def trace_gpu_usage(level: Optional[int] = None, show_if_disabed: Optional[bool] = False):
    """Trace out usage for GPU memory, etc.
    Note: This used nvidia-smi and is omitted for non-CUDA unless SHOW_IF_DISABED.
    The optional level defaults to GPU_DEBUG_LEVEL.

    TODO: support other types besides NVidia"""
    if level is None:
        level = GPU_DEBUG_LEVEL
    if ((TORCH_DEVICE == "cuda") or show_if_disabed):
        debug.trace(level, "GPU usage")
        debug.trace(level, gh.run("nvidia-smi"))
    else:
        debug.trace(level, "No CUDA device enabled, so skipping nvidia-smi")
    ## DEBUG: debug.trace_expr(1, int(level), int(DEBUG_LEVEL), GPU_DEBUG_LEVEL)
    if level > DEBUG_LEVEL:
        debug.trace(TL.USUAL, "FYI: Use higher trace level with trace_gpu_usage (DEBUG_LEVEL >= $level)")
 

#-------------------------------------------------------------------------------
        
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    system.print_error("Warning: Not intended for direct invocation.")
    debug.code(TL.USUAL, lambda: trace_gpu_usage())    # pylint: disable=unnecessary-lambda



#! /usr/bin/env python3
#
# Test(s) for ../gpu_utils.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_gpu_utils.py
#
#-------
# Sample tested output
#
# Sat Mar  2 17:27:08 2024       
# +---------------------------------------------------------------------------------------+
# | NVIDIA-SMI 535.98                 Driver Version: 535.98       CUDA Version: 12.2     |
# |-----------------------------------------+----------------------+----------------------+
# | GPU  Name                 Persistence-M | Bus-Id        Disp.A | Volatile Uncorr. ECC |
# | Fan  Temp   Perf          Pwr:Usage/Cap |         Memory-Usage | GPU-Util  Compute M. |
# |                                         |                      |               MIG M. |
# |=========================================+======================+======================|
# |   0  NVIDIA GeForce RTX 3080        Off | 00000000:01:00.0  On |                  N/A |
# |  0%   50C    P8              33W / 320W |   4786MiB / 10240MiB |      0%      Default |
# |                                         |                      |                  N/A |
# +-----------------------------------------+----------------------+----------------------+
# |   1  Quadro P400                    Off | 00000000:05:00.0  On |                  N/A |
# | 34%   43C    P0              N/A /  N/A |    130MiB /  2048MiB |      0%      Default |
# |                                         |                      |                  N/A |
# +-----------------------------------------+----------------------+----------------------+

"""Tests for gpu_utils module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
## TODO: from mezcla.unittest_wrapper import trap_exception
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system


# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
import mezcla.gpu_utils as THE_MODULE

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception         ## TODO: from mezcla import glue_helpers as gh
    # TODO: remove when debugged
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        output = self.run_script(options="", log_file=self.temp_file)
        log_contents = system.read_file(self.temp_file)
        self.do_assert(not output.strip())
        self.do_assert(my_re.search(r"not.*direct.*invocation", log_contents.strip().lower()))
        return

    @pytest.mark.skipif(not THE_MODULE.HAS_CUDA, reason="No CUDA support")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_gpu_usage(self):
        """TODO: flesh out test for whatever (capsys-like)"""
        debug.trace(4, f"TestIt2.test_02_whatever(); self={self}")
        THE_MODULE.trace_gpu_usage(level=1)
        captured = self.get_stderr()
        self.do_assert(my_re.search(r"NVIDIA-SMI\s+\S+\s+Driver Version:\s+\S+\s+CUDA.Version:\s+\S+",
                                    captured))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_type_hints(self):
        """Test out validation via pydantic"""
        # NOTE: This will not work until there is automatic support for pydantic type checking,
        # TODO2: write gpu_utils.py with validation decorators to to temp mezcla repo and invoke pytest using that repo
        # TODO?: mod="gpu_utils.py"; tmp_dir="/tmp/mezcla"; mkdir -p $tmp_dir $tmp_dir/tests; perl -pe 's/^def /\@valdate_call\n$&/;' $mod | perl -0777 -pe 's/^import/from pydantic import validate_call\n$&/;' >| $tmp_dir/$mod; copy-force tests/test_$mod $tmp_dir/tests/test_$mod; PYTHONPATH="$tmp_dir" pytest --runxfail $tmp_dir/tests/test_$mod
        debug.trace(4, f"TestIt2.test_03_type_hints(); self={self}")
        captured = self.get_stderr()
        try:
            THE_MODULE.trace_gpu_usage(level="two")
        except:
            pass
        assert("ValidationError" in captured)
    
#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../spacy_nlp.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_spacy_nlp.py
#
# Warning:
# - Maldito Spacy has some issues under Python 3.8 and 3.9, so the tests are
#   currently marked xfail until it stabilizes.
#
#...............................................................................
# Output from sample test:
#
# $ spacy_nlp.py - <<<$'I came, I saw, I conquered. The food\nwas bland.'
# text  is_oov  is_stop sentiment       is_sent_start
# It    False   True    0.0     True
# came  False   False   0.0     False
# ,     False   False   0.0     False
# it    False   True    0.0     False
# saw   False   False   0.0     False
# ,     False   False   0.0     False
# it    False   True    0.0     False
# conquered     False   False   0.0     False
# .     False   False   0.0     False
# text  is_oov  is_stop sentiment       is_sent_start
# The   False   True    0.0     True
# food  False   False   0.0     False
# was   False   True    0.0     False
# bland False   False   0.0     False
# .	False	False	0.0	False
#

"""Tests for spacy_nlp module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.misc_utils import is_close
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
## DEBUG: from mezcla.unittest_wrapper import trap_exception

# Environment manpulation
# note: needed to be done prior to loading Spacy (see THE_MODULE below)
# TODO2: rework to modification as part of test case (e.g., a la monkey patch)
system.setenv("SPACY_MODEL", "en_core_web_md")

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
try:
    import mezcla.spacy_nlp as THE_MODULE
except:
    THE_MODULE = None
    debug.trace_exception(1, "mezcla.spacy_nlp import")

class TestSentimentAnalyzer(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail
    def test_get_score(self):
        """Ensure SentimentAnalyzer.get_score works as expected"""
        debug.trace(4, "test_get_score()")
        sentiment = THE_MODULE.SentimentAnalyzer()
        self.do_assert(is_close(sentiment.get_score('bad'), -0.542))
        self.do_assert(is_close(sentiment.get_score('good'), 0.440))


class TestSpacyNlpUtils(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail
    ## TODO: @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_get_char_span(self):
        """Ensure get_char_span works as expected"""
        debug.trace(4, "test_get_char_span()")
        doc = THE_MODULE.SpacyHelper().nlp("the quick brown fox")
        # contract mode
        assert THE_MODULE.get_char_span(doc, 4, 19).text == "quick brown fox"
        assert THE_MODULE.get_char_span(doc, 4, 18).text == "quick brown"
        assert THE_MODULE.get_char_span(doc, 4, 20) is None

        # expand mode
        assert THE_MODULE.get_char_span(doc, 16, 16).text == "fox"
        assert THE_MODULE.get_char_span(doc, 14, 17).text == "brown fox"

        # adjusted span
        assert doc.text[15] == ' ' # make sure we're starting at a space
        assert THE_MODULE.get_char_span(doc, 15, 16).text == "fox"
        assert THE_MODULE.get_char_span(doc, 15, 15).text == "brown"


    @pytest.mark.xfail
    ## TODO: @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_pysbd_sentence_boundaries(self):
        """Ensure pysbd_sentence_boundaries works as expected"""
        debug.trace(4, "test_pysbd_sentence_boundaries()")
        self.do_assert(False, "TODO: code test")

    @pytest.mark.xfail
    ## TODO: @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_nltk_sentence_boundaries(self):
        """Ensure nltk_sentence_boundaries works as expected"""
        debug.trace(4, "test_nltk_sentence_boundaries()")
        self.do_assert(False, "TODO: code test")


class TestSpacy(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail
    def test_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        data = ["It came, it saw, it conquered. The food", "was bland."]
        system.write_lines(self.temp_file, data)
        output = self.run_script(options="", data_file=self.temp_file)
        actual = output.strip()
        debug.trace_expr(5, actual, max_len=1024)
        # Check for header
        self.do_assert(my_re.search(r"text\tis_oov\tis_stop\tsentiment\tis_sent_start", actual))
        # Check for start of sentences
        self.do_assert(my_re.search(r"It\tFalse\tTrue\t0.0\tTrue", actual))
        self.do_assert(my_re.search(r"The\tFalse\tTrue\t0.0\tTrue", actual))
        # Check that no other sentence starts (n.b., "was" doesn't even though new line)
        self.do_assert(my_re.search(r"was\tFalse\tTrue\t0.0\tFalse", actual))
        sent_start_info = sorted(gh.extract_matches(r"\t(\S+)$", actual.splitlines()))
        self.do_assert(sent_start_info[-5:] == ["False", "True", "True", "is_sent_start", "is_sent_start"])
        return

    @pytest.mark.xfail
    def test_chunker(self):
        """Test NP chunking"""
        debug.trace(4, f"TestIt2.test_chunker(); self={self}")
        chunker = THE_MODULE.Chunker()
        text = "my dog has fleas"
        expected_NPs = ["my dog", "fleas"]
        actual_NPs = chunker.noun_phrases(text)
        self.do_assert(expected_NPs == actual_NPs)
        return


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# tpo_common.py: Python module with common utilities mostly for debugging,
# mostly superceded by debug.py, system.py, and misc_util.py """
#
# sample usage:
#    from mezcla import tpo_common as tpo
#    tpo.debug_print("start: " + tpo.debug_timestamp(), 3)
#
#------------------------------------------------------------------------
#
# Note:
# - *** Obsolete: use system.py, debug.py, etc. instead.
# - Use Emacs's query-replace-regexp (e.g., via M-#) to interative convert debug_format to debug.trace_fmt: 
#     from: debug_format(\(\"[^\"]*\"\), \([0-9]*\),
#       to: debug.trace_fmt(\2, \1,")
# - The TPO prefix was used originally to avoid conflicts. Some of the older scripts
#   rely upon this, but it will be eventually phased out.
# - This redefines some built-in functions (e.g., format and exit), so
#   you shouldn't import using 'import *'.
# - Debugging output and error messages are converted to UTF8 before output
#   via s=unicode(s, "utf8"), which is same as s=codecs.encode(s, "utf8").
# - Debugging code is "conditionally compiled" to avoid overhead of passing
#   arguments to functions that do nothing in release code.
# - Includes support for overriding stderr (e.g., for use with embedded apps as in adhoc/optimize_company_extraction.py's invocation of extract_company_info.py).
#
# TODO1:
# - *** Make sure any function not mapping directly to debug or system traces out
#   with "tpo" for grep purposes (e.g.,  tpo.debug_format in debug_format). ***
#   This will help flag functions to be ported (e.g., redirect_stderr).
#
# TODO:
# - Change all functions to using docstrings for comments already!!!
# - Add support for using pprint module
# - Put a version of assertion() here (i.e., from glue_helpers.py).
# - Extend getenv support to indicate user-vs-devel options.
# - Remove obsolete functions (e.g., get_current_function_name).
# - Use debug_trace in place of debug_print in code run often.
#
#------------------------------------------------------------------------
#

"""Common utility functions"""

# Load required libraries
#
# - For Debugging timestamp purposes (e.g., OUTPUT_DEBUG_TIMESTAMPS support)
## OLD:
## from datetime import datetime
## import atexit
#
# - The usuals:
import sys
import os
import re
# - Others
import inspect
import logging
import pickle
from six import string_types
## OLD: import time
## OLD: import types

# Load OrderedDict (added in Python 2.7)                                       
try:
    from collections import OrderedDict
except ImportError:
    try:
        from ordereddict import OrderedDict
    except ImportError:
        def OrderedDict():
            """Legacy stub"""
            assert False, "requires ordereddict package (or python 2.7+)"

# Local packages
from mezcla import debug
from mezcla.debug import trace_level as debug_level
from mezcla import system
            
# Defaults for globals
## OLD:
## debug_level = 0                          # level at which debug tracing occurs
## output_timestamps = False                # whether to prefix output with timestamp
## use_logging = False                      # traces via logging facility (and stderr)
skip_format_warning = False              # skip warning about format w/o namespace
stderr = sys.stderr                      # file handle for error messages 

# Constants for use with debug_print, etc.
LEVEL1 = 1
LEVEL2 = 2
LEVEL3 = 3
LEVEL4 = 4
LEVEL5 = 5
LEVEL6 = 6
# Aliases for debugging levels (by convention)
ALWAYS = 0
CRITICAL = ALWAYS
ERROR = 1
WARNING = 2
USUAL = 3
DETAILED = 4
VERBOSE = 5
QUITE_DETAILED = 6
QUITE_VERBOSE = 7
DEFAULT_DEBUG_LEVEL = ALWAYS

# Other constants
# TODO: use sys.version_info.major, etc.
USE_SIMPLE_FORMAT = (sys.version_info[0] <= 2) and (sys.version_info[1] <= 5)

#------------------------------------------------------------------------
# Debugging functions
#
# Notes:
# - These are no-op's unless __debug__ is True.
# - Running python with the -O (optimized) option ensures that __debug__ is False.
#

## OLD:
## if __debug__:
## 
##     def set_debug_level(level):
##         """Set new debugging LEVEL"""
##         global debug_level
##         debug_level = level
##         return
## 
## 
##     def debugging_level():
##         """Get current debugging level"""
##         global debug_level
##         return debug_level
## 
## 
##     def debug_trace_without_newline(text, *args, **kwargs):
##         """Print TEXT without trailing newline, provided at debug trace LEVEL or higher, using ARGS for %-placeholders
##         Notes: ensures text encoded as UTF8 if under Python 2.x;
##         also, exceptions are ignored (to encourage more tracing)."""
##         # Warning: To avoid recursion don't call other user functions here unless
##         # they don't uses tracing (e.g., _normalize_unicode).
##         # TODO: work out shorter name (e.g., debug_trace_no_eol)
##         #
##         global debug_level
##         level = kwargs.get('level', 1)
##         if debug_level >= level:
##             if debug_level >= 96:
##                 stderr.write("debug_trace_without_newline(text=%s, level=%s, args=%s)" % 
##                              (_normalize_unicode(text), level, [_normalize_unicode(v) for v in args]))
##             if args:
##                 try:
##                     text = (text % args)
##                 except (AttributeError, IndexError, NameError, TypeError, ValueError):
##                     print_stderr("Exception in debug_print: " + str(sys.exc_info()))
##                     
##             # Output optional timestamp (e.g., for quick-n-dirty profiling)
##             if output_timestamps:
##                 # Get time-proper from timestamp (TODO: find standard way to do this)
##                 # TODO: make date-stripping optional
##                 timestamp = re.sub(r"^\d+-\d+-\d+\s*", "", debug_timestamp())
##                 stderr.write("[%s] " % timestamp)
##             # Output text, making sure text represented in UTF8 if needed (n.b., via inlined ensure_unicode)
##             text = _normalize_unicode(text)
##             stderr.write(text)
##             if use_logging:
##                 logging.debug(text)
##             ##
##             ## # Sanity check to ensure no instantiatable templates used in text
##             ## # TODO: rework debug_print to safely handle templates by default
##             ## while True:
##             ##     #                   +-pre--++---variable----++--rest-----+
##             ##     match = re.search(r"(^|[^{])({[A-Za-z0-9_]+})(([^}]|$|).*)", text)
##             ##     if not match:
##             ##         break
##             ##     text = match.group(3)
##             ##     var_format = match.group(2)
##             ##     if format(var_format, True, True) != var_format:
##             ##         print_stderr("Warning: template used with instantiated variable; try debug_format instead")
##         return
## 
##     def debug_trace(text, *args, **kwargs):
##         """Prints TEXT (formatted with ARG1, ...) if at LEVEL or higher.
##         Note: To avoid needless evaluation arguments should be specified 
##         as distinct paramaters rather than using string format operator (%)."""
##         # Note: Implemented in terms of debug_trace_without_newline to keep timestamp support in one place.
##         # TODO: add skip_newline option
##         global debug_level
##         debug_trace_without_newline(text, *args, **kwargs)
##         level = kwargs.get('level', 1)
##         # TODO: assertion(isinstance(level, int))
##         if debug_level >= level:
##             stderr.write("\n")
##         return
## 
##     def debug_print_without_newline(text, level=1):
##         """Wrapper around debug_trace_without_newline (q.v.)
##         Note: Consider using debug_trace_without_newline directly."""
##         return debug_trace_without_newline(text, level=level)
## 
##     def debug_print(text, level=1, skip_newline=False):
##         """Print TEXT if at debug trace LEVEL or higher.
##         Notes:
##         - Implemented in terms of debug_print_without_newline to keep timestamp support in one place.
##         - Consider using debug_trace instead (to avoid string interpolation overhead)."""
##         # TODO: assertion(isinstance(level, int))
##         global debug_level
##         debug_trace_without_newline(text, level=level)
##         if (debug_level >= level) and (not skip_newline):
##             stderr.write("\n")
##         return
## 
## 
##     def debug_format(text, level=1, skip_newline=False, **namespace):
##         """Version of debug_print that expands TEXT using format.
##         Note: Exceptions are ignored (to encourage tracing, not discourage)."
##         """
##         # NOTE: String values from namespace need to be in UTF-8 format.
##         # TODO: rename as debug_print_format as not just formatting the text
##         global debug_level
##         assert(isinstance(level, int))
##         if debug_level >= level:
##             ignore_exc = (debug_level < QUITE_DETAILED)
##             debug_print(format(text, indirect_caller=True, 
##                                ignore_exception=ignore_exc, **namespace),
##                         skip_newline=skip_newline)
##         return
## 
## 
##     def debug_timestamp():
##         """Return timestamp for use in debugging traces"""
##         # EX: debug_timestamp() => "2015-01-18 16:39:45.224768"
##         # TODO: use format compatible with logging (e.g., comma in place of period before micrososeconds)
##         return to_string(datetime.now())
## 
## 
##     def debug_raise(level=1):
##         """Raise an exception if debugging at specified trace LEVEL or higher"""
##         # Note: Intended for use in except clause to produce full stacktrace when debugging.
##         # TODO: Have version that just prints complete stacktrace (i.e., without breaking).
##         if debug_level >= level:
##             raise                        # pylint: disable=misplaced-bare-raise
##         return
## 
## 
##     def trace_array(array, level=VERBOSE, label=None):
##         """Output the (list) array to STDERR if debugging at specified trace LEVEL or higher, using LABEL as prefix"""
##         global debug_level
##         assert(isinstance(level, int))
##         if debug_level >= level:
##             trace_output = ("%s: " % label) if label else ""
##             for i, item in enumerate(array):
##                 trace_output += "  " if (i > 0) else ""
##                 trace_output += "%d: %s" % (i, item)
##             debug_print(trace_output)
##         return
## 
## 
##     def trace_object(obj, level=VERBOSE, label=None, show_private=None, show_methods_etc=None, indent="", max_value_len=1024):
##         """Traces out OBJ instance to stderr if at debugging LEVEL or higher, 
##         optionally preceded by LABEL, using INDENT, and limiting value lengths to MAX_VALUE_LEN"""
##         # based on http://stackoverflow.com/questions/192109/is-there-a-function-in-python-to-print-all-the-current-properties-and-values-of
##         debug_format("trace_object(_, lvl={lvl}, lab={lab}, prv={prv}, mth={mth}, ind={ind})", 7,
##                      lvl=level, lab=label, prv=show_private, mth=show_methods_etc, ind=indent)
##         global debug_level
##         assert(isinstance(level, int))
##         if debug_level >= level:
##             debug_print("%s%s: {" % (indent, label if label else obj))
##             if show_private is None:
##                 show_private = (level >= 6)
##             if show_methods_etc is None:
##                 show_methods_etc = (level >= 7)
##             for attr in dir(obj):
##                 # Note: unable to filter properties directly (so try/except handling added)
##                 # See Take1 and Take2 below.
##                 try:
##                     ## TAKE2:
##                     ## if re.search("property|\?\?\?", str(getattr(type(obj), attr, "???"))):
##                     ##     debug_print("%s: property" % attr, 7)
##                     ##     continue
##                     # note: omits internal variables unless at debug level 6+; also omits methods unless at debug level 7+
##                     debug_print("%s\tattr: name: %s" % (indent, attr), 8)
##                     if (attr[0] != '_') or show_private:
##                         attr_value = getattr(obj, attr)
##                         debug_trace("%s\tattr: type: %s",
##                                     indent, type(attr_value), level=8)
##                         method_types = (
##                             types.MethodType, types.FunctionType, 
##                             types.BuiltinFunctionType, types.BuiltinMethodType,
##                             types.ModuleType)
##                         if ((not isinstance(attr_value, method_types))
##                                 or ("method-wrapper" not in str(type(attr_value)))
##                                 or show_methods_etc):
##                             if len(to_string(attr_value)) > max_value_len:
##                                 attr_value = to_string(attr_value)[:max_value_len] + "..."
##                             debug_print("%s\t%s = %r" % (indent, attr, attr_value))
##                 except (KeyError, ValueError, AttributeError):
##                     debug_print("%s\t%s = ???" % (indent, attr))
##             debug_print("%s}" % indent)
##         return
## 
## 
##     def trace_value(value, level=5, label=None):
##         """Traces out VALUE to stderr if at debugging LEVEL or higher with optional '<LABEL>: ' prefix.
##         @Note: If value is an array or dict, each entry is output on a separate line."""
##         global debug_level
##         assert(isinstance(level, int))
##         if debug_level >= level:
##             prefix = ""
##             value_spec = value
##             if label:
##                 prefix = ("%s: " % label)
##             if isinstance(value, list):
##                 value_spec = "{\n" + "\n".join(["\t%s: %s" % (i, v) for (i, v) in enumerate(value)]) + "}\n"
##             elif isinstance(value, dict):
##                 value_spec = "{\n"
##                 for key in value.keys():
##                     # Note: Exception handling used in case value access leads to error,
##                     # such as with hash tables having Mako placeholders for undefined values.
##                     try:
##                         key_value = value[key]
##                         debug_print("key=%s key_value=%s" % (key, key_value), 9)
##                     except NameError:
##                         print_stderr("Exception in trace_value getting value for key %s" % key)
##                         key_value = "n/a"
##                     value_spec += "\t%s: %s\n" % (key, key_value)
##                 value_spec += "}\n"
##             debug_print("%s%s" % (prefix, value_spec))
##         return
## 
## 
##     def trace_current_context(level=QUITE_DETAILED, label=None, 
##                               show_methods_etc=False):
##         """Traces out current context (local and global variables), with output
##         prefixed by "LABEL context" (e.g., "current context: {\nglobals: ...}").
##         Notes: By default the debugging level must be quite-detailed (6).
##         If the debugging level is higher, the entire stack frame is traced.
##         Also, methods are omitted by default."""
##         frame = None
##         if label is None:
##             label = "current"
##         try:
##             frame = inspect.currentframe().f_back
##         except (AttributeError, KeyError, ValueError):
##             debug_print("Exception during trace_current_context: " + 
##                         to_string(sys.exc_info()), 5)
##         debug_format("{label} context: {{", level, label=label)
##         prefix = "    "
##         if (debugging_level() - level) > 1:
##             trace_object(frame, (level + 2), "frame", indent=prefix,
##                          show_methods_etc=show_methods_etc)
##         else:
##             debug_format("frame = {f}", level, f=frame)
##             if frame:
##                 trace_object(frame.f_globals, level, "globals", indent=prefix,
##                              show_methods_etc=show_methods_etc)
##                 trace_object(frame.f_locals, level, "locals", indent=prefix,
##                              show_methods_etc=show_methods_etc)
##         debug_trace("}", level=level)
##         return
## 
## 
##     def during_debugging(expression=True):
##         """Returns True if debugging, optionally conditioned upon EXPRESSION
##          Note: The expression is not considered in non-debug mode"""
##         return (debugging_level() > 0) and expression
## 
## else:
## 
##     def set_debug_level(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def debugging_level(*_args, **_kwargs):
##         """Non-debug stub"""
##         return 0
## 
## 
##     def debug_trace_without_newline(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def debug_trace(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def debug_print_without_newline(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def debug_print(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def debug_format(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def debug_timestamp(*_args, **_kwargs):
##         """Non-debug stub"""
##         return ""
## 
## 
##     def debug_raise(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def trace_array(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def trace_object(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def trace_value(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def trace_current_context(*_args, **_kwargs):
##         """Non-debug stub"""
##         pass
## 
## 
##     def during_debugging(*_args, **_kwargs):
##         """Non-debug stub"""
##         return False
## 

def set_debug_level(level):
    """Set new debugging LEVEL"""
    return debug.set_level(level)


def debugging_level():
    """Get current debugging level"""
    return debug.get_level()


def debug_trace_without_newline(text, *args, **kwargs):
    """Print TEXT without trailing newline, provided at debug trace LEVEL or higher, using ARGS for %-placeholders
    Notes: ensures text encoded as UTF8 if under Python 2.x;
    also, exceptions are ignored (to encourage more tracing)."""
    level = kwargs.get('level', 1)
    if args:
        try:
            text = (text % args)
        except (AttributeError, IndexError, NameError, TypeError, ValueError):
            system.print_exception_info("debug_trace_without_newline")
    return debug.trace(level, text, no_eol=True)


def debug_trace(text, *args, **kwargs):
    """Prints TEXT (formatted with ARG1, ...) if at LEVEL or higher.
    Note: To avoid needless evaluation arguments should be specified 
    as distinct paramaters rather than using string format operator (%)."""
    return debug_trace_without_newline(text + "\n", *args, **kwargs)


def debug_print_without_newline(text, level=1):
    """Wrapper around debug_trace_without_newline (q.v.)
    Note: Consider using debug_trace_without_newline directly."""
    return debug_trace_without_newline(text, level=level)


def debug_print(text, level=1, skip_newline=False):
    """Print TEXT if at debug trace LEVEL or higher.
    Notes:
    - Implemented in terms of debug_print_without_newline to keep timestamp support in one place.
    - Consider using debug_trace instead (to avoid string interpolation overhead)."""
    # TODO: assertion(isinstance(level, int))
    out_text = (text if skip_newline else (text + "\n"))
    return debug_trace_without_newline(out_text, level=level)


def debug_format(text, level=1, skip_newline=False, **namespace):
    """Version of debug_print that expands TEXT using format.
    Note: Exceptions are ignored (to encourage tracing, not discourage)."
    """
    debug.trace(6, f"tpo.debug_format({text}, {skip_newline}, ...)",
                skip_sanity_checks=True)
    ## TODO: return debug.trace(no_eol=skip_newline, level, text, **namespace)
    # NOTE: String values from namespace need to be in UTF-8 format.
    # TODO: rename as debug_print_format as not just formatting the text
    assert(isinstance(level, int))
    if (debugging_level() >= level):
        ignore_exc = (debugging_level() < QUITE_DETAILED)
        debug_print(format(text, indirect_caller=True, 
                           ignore_exception=ignore_exc, **namespace),
                    skip_newline=skip_newline)
    return


def debug_timestamp():
    """Return timestamp for use in debugging traces"""
    # EX: debug_timestamp() => "2015-01-18 16:39:45.224768"
    # TODO: use format compatible with logging (e.g., comma in place of period before micrososeconds)
    return debug.timestamp()


def debug_raise(level=1):
    """Raise an exception if debugging at specified trace LEVEL or higher"""
    # Note: Intended for use in except clause to produce full stacktrace when debugging.
    # TODO: Have version that just prints complete stacktrace (i.e., without breaking).
    return debug.raise_exception(level)


def trace_array(array, level=VERBOSE, label=None):
    """Output the (list) array to STDERR if debugging at specified trace LEVEL or higher, using LABEL as prefix"""
    return debug.trace_values(level, array, label)


def trace_object(obj, level=VERBOSE, label=None, show_private=None, show_methods_etc=None, indent="", max_value_len=1024):
    """Traces out OBJ instance to stderr if at debugging LEVEL or higher, 
    optionally preceded by LABEL, using INDENT, and limiting value lengths to MAX_VALUE_LEN"""
    return debug.trace_object(level, obj, label, show_private=show_private, show_methods_etc=show_methods_etc, indentation=indent, max_value_len=max_value_len)


def trace_value(value, level=5, label=None):
    """Traces out VALUE to stderr if at debugging LEVEL or higher with optional '<LABEL>: ' prefix.
    @Note: If value is an array or dict, each entry is output on a separate line."""
    assert(isinstance(level, int))
    if (debugging_level() >= level):
        prefix = ""
        value_spec = value
        if label:
            prefix = ("%s: " % label)
        if isinstance(value, list):
            value_spec = "{\n" + "\n".join(["\t%s: %s" % (i, v) for (i, v) in enumerate(value)]) + "}\n"
        elif isinstance(value, dict):
            value_spec = "{\n"
            for key in value.keys():
                # Note: Exception handling used in case value access leads to error,
                # such as with hash tables having Mako placeholders for undefined values.
                try:
                    key_value = value[key]
                    debug_print("key=%s key_value=%s" % (key, key_value), 9)
                except NameError:
                    print_stderr("Exception in trace_value getting value for key %s" % key)
                    key_value = "n/a"
                value_spec += "\t%s: %s\n" % (key, key_value)
            value_spec += "}\n"
        debug_print("%s%s" % (prefix, value_spec))
    return


def trace_current_context(level=QUITE_DETAILED, label=None, 
                          show_methods_etc=False, **kwargs):
    """Traces out current context (local and global variables), with output
    prefixed by "LABEL context" (e.g., "current context: {\nglobals: ...}").
    Notes: By default the debugging level must be quite-detailed (6).
    If the debugging level is higher, the entire stack frame is traced.
    Also, methods are omitted by default."""
    return debug.trace_current_context(level=level, label=label, show_methods_etc=show_methods_etc, indirect=True, **kwargs)


def during_debugging(expression=True):
    """Returns True if debugging, optionally conditioned upon EXPRESSION
     Note: The expression is not considered in non-debug mode"""
    return (debugging_level() > 0) and expression


def debugging(level=(DEFAULT_DEBUG_LEVEL + 1)):
    """Whether running with LEVEL or higher debugging (1 by default)"""
    return debugging_level() >= level


def detailed_debugging():
    """Whether running with detailed debug tracing"""
    return debugging_level() >= DETAILED


def verbose_debugging():
    """Whether running with verbose debug tracing"""
    return debugging_level() >= VERBOSE


#------------------------------------------------------------------------
# General utility functions

def to_string(obj):
    """Returns string rendition of OBJ (i.e., %s-based)"""
    # Note: str() not used in order to properly handle unicode
    # EX: to_string(123) => "123"
    # EX: to_string(u"\u1234") => u"\u1234"
    # EX: to_string(None) => "None"
    # TODO: have specialization that returns "" for None
    result = obj
    ## NOTE: Gotta hate python
    ## if not isinstance(result, types.StringTypes):
    if not isinstance(result, string_types):
        result = "%s" % (obj,)
    return result

def normalize_unicode(text, encoding="utf8"):
    """Converts Unicode TEXT to ENCODING (normally UTF-8) if necessary, such as
    to prepare for output. Note: this is a no-op in Python 3 or higher"""
    result = _normalize_unicode(text, encoding)
    debug_format("normalize_unicode({t}, {e}) => {r}", 10, 
                 t=text, e=encoding, r=result)
    debug_format("\ttypes: in={it} out={ot}", 11, 
                 it=type(text), ot=type(result))
    return result
#
def _normalize_unicode(text, encoding="utf8"):
    """Implementation of normalize_unicode (q.v.)"""
    # EX: normalize_unicode(u'\u1234') => '\xe1\x88\xb4'
    # Note: As this is used in suport for debug_print, no tracing is done.
    # TODO: work out more-untuitive name
    result = text
    if sys.version_info[0] < 3:
        if isinstance(result, unicode):     # pylint: disable=undefined-variable
            result = result.encode(encoding)
    return result


def ensure_unicode(text, encoding="utf8"):
    """Ensures TEXT is encoded as unicode, using ENCODING (e.g., UTF8). Note: this is a no-op in Python 3 or higher"""
    # EX: ensure_unicode('\xe1\x88\xb4') => u'\u1234'
    result = _ensure_unicode(text, encoding)
    debug_format("ensure_unicode({t}) => {r}", 10, 
                 t=text, r=result)
    debug_format("\ttypes: in={it} out={ot}", 11, 
                 it=type(text), ot=type(result))
    return result
#
def _ensure_unicode(text, encoding="utf8"):
    """Implementation of ensure_unicode (q.v.)"""
    # Note: As this used in support for debug_print, no tracing is done.
    result = text
    if sys.version_info[0] < 3:
        # pylint: disable=undefined-variable
        if not isinstance(result, unicode):
            result = unicode(result, encoding)
    return result


def print_stderr(text, **namespace):
    """Output TEXT to standard error (with newline added), using optional
    NAMESPACE for format"""
    # Note: ensures text encoded as UTF8 if under Python 2.x
    if namespace:
        text = format(text, **namespace)
    stderr.write(normalize_unicode(text) + "\n")
    return


def redirect_stderr(filename):
    """Redirects error output to FILENAME"""
    global stderr
    assert(stderr == sys.stderr)
    stderr = system.open_file(filename, "w")
    assert(stderr)


def restore_stderr():
    """Restores error output to system stderr, closing handle for redirection"""
    global stderr
    assert(stderr != sys.stderr)
    if stderr:
        stderr.close()
    stderr = sys.stderr


def exit(message, **namespace):    # pylint: disable=redefined-builtin
    """Display error MESSAGE to stderr and then exit, using optional
    NAMESPACE for format"""
    if namespace:
        message = format(message, **namespace)
    print_stderr(message + "\n")
    return sys.exit()


def setenv(var, value):
    """Set environment VAR to VALUE"""
    debug_print("setenv('%s', '%s')" % (var, value), 6)
    os.environ[var] = to_string(value) if value is not None else ""
    return


def chomp(text, line_separator=os.linesep):
    """Removes trailing occurrence of LINE_SEPARATOR from TEXT"""
    # EX: chomp("abc\n") => "abc"
    # EX: chomp("http://localhost/", "/") => "http://localhost"
    result = text
    if result.endswith(line_separator):
        new_len = len(result) - len(line_separator)
        result = result[:new_len]
    debug_format("chomp({t}, {sep}) => {r}", 8,
                 t=text, sep=line_separator, r=result)
    return result

#-------------------------------------------------------------------------------
# Environment accessors and setters

def getenv(var):
    """Return value of environment VAR"""
    value = os.getenv(var)
    debug_print("getenv('%s') => '%s'" % (var, value), 8)
    return value


env_options = {}
env_defaults = {}
#
def register_env_option(var, description, default):
    """Register environment VAR as option with DESCRIPTION and DEFAULT"""
    debug_format("register_env_option({v}, {d})", 7, v=var, d=description)
    global env_options
    global env_defaults
    env_options[var] = (description or "")
    env_defaults[var] = default
    return


def get_registered_env_options():
    """Returns list of environment options registered via register_env_option"""
    ## OLD: option_names = [k for k in env_options if (env_options[k] and env_options[k].strip())]
    option_names = [k for (k, v) in env_options.items() if (v and v.strip())]
    ## OLD: debug_format("get_registered_env_options() => {option_names}", 5)
    debug_print(f"get_registered_env_options() => {option_names}", 5)
    return option_names


def get_environment_option_descriptions(include_all=None, include_default=None, indent=" "):
    """Returns list of environment options and their descriptions"""
    debug_format("env_options={eo}", 5, eo=env_options)
    debug_format("env_defaults={ed}", 5, ed=env_defaults)
    if include_all is None:
        include_all = verbose_debugging()
    if include_default is None:
        include_default = True
    #
    def _format_env_option(opt):
        """Returns OPT description and optionally default value (if INCLUDE_DEFAULT)"""
        debug_format("_format_env_option({opt})", 7)
        desc_spec = env_options.get(opt, "_")
        default_spec = ""
        if include_default:
            default_value = env_defaults.get(opt, None)
            has_default = (default_value is not None)
            default_spec = ("(%s)" % default_value) if has_default else "n/a"
        default_spec = default_spec.replace("\n", "\\n")
        return (opt, desc_spec + indent + default_spec)
    #
    ## OLD: option_descriptions = [_format_env_option(opt) for opt in env_options if (env_options[opt] or include_all)]
    option_descriptions = [_format_env_option(opt) for (opt, v) in env_options.items() if (v or include_all)]
    debug_format("get_environment_option_descriptions() => {od}", 5,
                 od=option_descriptions)
    return option_descriptions


def formatted_environment_option_descriptions(sort=False, include_all=None, indent="\t"):
    """Returns string list of environment options and their descriptions (separated by newlines and tabs), optionally SORTED"""
    option_info = get_environment_option_descriptions(include_all)
    if sort:
        option_info = sorted(option_info)
    entry_separator = "\n%s" % indent
    descriptions = entry_separator.join(["%s%s%s" % (opt, indent, (desc if desc else "n/a")) for (opt, desc) in option_info])
    debug_format("formatted_environment_option_descriptions() => {d}", 6,
                 d=descriptions)
    return descriptions


def getenv_value(var, default=None, description=None, export=None):
    """Returns textual value for environment variable VAR; if not set, uses DEFAULT which can be None. If DESCRIPTION given, the variable is added to list of registered options. If EXPORT specified, the environment entry will be created if necessary (useful during sub-script invocation)."""
    # Note: Use one of the specialized variants (e.g., getenv_text, getenv_boolean, or getenv_number).
    value = getenv(var)
    if export is None:
        ## OLD: export = description is not None
        export = False
    if not value:
        debug_print("getenv_value: no value for %s" % var, 7)
        value = default
        if export:
            setenv(var, value)
    register_env_option(var, description, default)
    debug_print("getenv_value('%s') => %s" % (var, value), 6)
    return value


def getenv_text(var, default="", description=None):
    """Returns textual value for environment variable VAR (or string version of DEFAULT unless None)"""
    # Note: This is a simple wrapper around getenv_value using lower tracing level (as former is intended as helper for specialized types given below).
    # TODO: define in terms of system.getenv_text
    value = getenv_value(var, default, description=description)
    text_value = to_string(value) if value is not None else ""
    debug_print("getenv_text('%s', '%s') => '%s'" % (var, default, text_value), 4)
    return text_value


def getenv_bool(var, default=False, description=None):
    """Returns boolean flag based on environment VAR (or DEFAULT value which can be None). Note: "0" or "False" is interpreted as False, and any value as True."""
    bool_value = default
    value_text = getenv_text(var, default, description=description)
    # TODO: gh.assertion(type(default) == bool or default is None)
    if ((default is not None) and (not isinstance(default, bool))):
        print_stderr("Warning: unexpected default type to getenv_boolean: " +
                     "{t} vs. bool or None; var={v}", t=type(default), v=var)
    # TODO: assert(isinstance(value_text, str))
    if value_text:
        bool_value = True
        if (value_text.lower() == "false") or (value_text == "0"):
            bool_value = False
    debug_print("getenv_boolean(%s, %s) => %s" % (var, default, bool_value), 4)
    return bool_value
#
getenv_boolean = getenv_bool


def getenv_number(var, default=-1, description=None, integral=False):
    """Returns number based on environment VAR (or DEFAULT value which can be None)"""
    num_value = default
    value_text = getenv_text(var, default, description=description)
    # TODO: assert(isinstance(value_text, str))
    # TODO: make sure misc_utils version reconciled (and rework tpo_common to use latter!)
    if value_text:
        # Note: safe_int/_float not used so that exception into always displayed
        try:
            num_value = int(value_text) if integral else float(value_text)
        except:
            print_stderr(format("Exception converting env number for {v} ('{val}'): {exc}",
                                v=var, val=value_text, exc=str(sys.exc_info())))
    debug_print("getenv_number(%s, %s) => %s" % (var, default, num_value), 4)
    return num_value


def getenv_int(var, default=-1, description=None):
    """Variant of getenv_number for integers. Note: returns integer unless None is default."""
    return getenv_number(var, default, description=description, integral=True)
#
getenv_integer = getenv_int


def getenv_real(var, default=-1, description=None):
    """Variant of getenv_number for reals (i.e., floating point). Note: (returns float unless None is default)."""
    return getenv_number(var, default, description=description, integral=False)


def getenv_float(var, default=-1, description=None):
    """Alias for getenv_real"""
    return getenv_number(var, default, description=description)


def get_current_function_name():
    """Returns the name of the currently executing function (excluding this function of course)"""
    # Note: Based on http://code.activestate.com/recipes/66062-determining-current-function-name.
    # Also see http://docs.python.org/2/reference/datamodel.html and http://docs.python.org/2/library/inspect.html.
    # TODO: remove as no longer used???
    name = "???"
    try:
        name = sys._getframe(1).f_code.co_name     # pylint: disable=protected-access
    except (AttributeError, ValueError):
        print_stderr("Exception in get_current_function_name: " + str(sys.exc_info()))
    debug_print("get_current_function_name() => %s" % name, 5)
    return name


def get_property_value(obj, property_name, default_value=None):
    """Gets property value for NAME from OBJ, using DEFAULT"""
    # EX: import datetime; (get_property_value(datetime.date.today(), 'year', -1) >= 2012) => True
    # EX: get_property_value(datetime.date.today(), 'minute', -1) => -1
    value = default_value
    if hasattr(obj, property_name):
        value = getattr(obj, property_name)
    debug_print("get_property_value%s => %s" % (to_string((obj, property_name, default_value)),
                                                value), level=4)
    return value


def simple_format(text, namespace):
    """Resolve format-style command argument specifications within TEXT (within angle brackets), using bindings from NAMESPACE"""
    # Note: This only supports evaluation of the text within braces (e.g., format specifiers not handled)
    result = text

    # Replace all placeholders via eval of expression
    while (True):
        # Find next variable reference (uses lookbehind to exclude double braces)
        #                  +no brace+ {expressn} +
        match = re.search("(?:[^{]|^)({([^{}]+)})", result)
        if not match: 
            break
        pattern = match.group(1)
        var = match.group(2)
        # Determine value and update argument
        value = None
        try:
            # note: no global namespace specified (but should be part of local one as with format below)
            value = eval(var, None, namespace)    # pylint: disable=eval-used
        except (IndexError, NameError, SyntaxError, TypeError, ValueError):
            debug_print("Exception during eval: " + str(sys.exc_info()), 6)
        if value is None:
            debug_print("Unable to resolve replacement '%s' in '%s'" % (var, result), 2)
            break
        result = result.replace(pattern, str(value))

    # Handle common escapes
    result = result.replace("{{", "{")
    result = result.replace("}}", "}")

    debug_print("simple_format('%s', [namespace]) => %s" % (text, result), 6)
    debug_print("\tnamespace: %s" % namespace, 7)
    return result


warned_about_namespace = {}
#
def format(text, indirect_caller=False, ignore_exception=False, **namespace):    # pylint: disable=redefined-builtin
    """Formats TEXT using local namespace, optionally using additional level or
    indirection. If no keywords NAMESPACE specified, they are taken from local
    environment (which is convenient but un-pythonic and a bit inefficient).
    Exceptions can be ignored (to support debug_format)."""
    # Notes:
    # - Argments need to be in UTF-8 if the template text is an ascii string.
    # - The technique for resolving the locals is based on Stack Overflow:
    #   http://stackoverflow.com/questions/6618795/get-locals-from-calling-namespace-in-python
    # - This conflicts with the format builtin, but that is only a potential
    #   issue if immported via 'from tpo_common import *'. The standard function
    #   can be accessed via __builtin__.format (or via builtins under Python 3).
    # - Encodes text in unicode prior to format proper and then encodes result
    #   into UTF-8.
    # Warning: don't call other user functions except debug_print or helper
    # functions that don't trace (e.g., _ensure_unicode).
    if (debugging_level() >= 99):
        debug_trace("format(%s, %s, %s, %s)",
                    text, indirect_caller, ignore_exception, namespace)
    # TODO: make unicode conversion optional
    ## OLD: text = _ensure_unicode(text)
    frame = None
    result = ""
    try:
        if not namespace:
            frame = inspect.currentframe().f_back
            if indirect_caller:
                frame = frame.f_back
            if not skip_format_warning:
                filename = frame.f_globals.get("__file__", "???")
                if filename.endswith(".pyc"):
                    filename = filename[:-1]
                source_line_key = "{f}:{l}".format(f=filename, l=frame.f_lineno)
                if source_line_key not in warned_about_namespace:
                    debug_trace("Warning: deriving namespace from locals and " +
                                "globals at %s: inefficient and un-pythonic!",
                                source_line_key, level=WARNING)
                    warned_about_namespace[source_line_key] = True
            trace_object(frame, 9, "frame")
            namespace = frame.f_globals.copy()
            namespace.update(frame.f_locals)
        else:
            namespace = namespace.copy()
        if USE_SIMPLE_FORMAT:
            result = simple_format(text, namespace)
        else:
            # Make sure values to output are UTF-8 unless text is unicode
            # Note: only converts strngs, lists, and hashes (e.g., omitting user objects and numbers)
            if sys.version_info[0] < 3:
                debug_trace("type(text)=%s", type(text), level=91)
                debug_trace("namespace=%s", namespace, level=94)
                ## OLD: transform_fn = _normalize_unicode if not type(text) == unicode else _ensure_unicode
                # pylint: disable=undefined-variable
                transform_fn = _normalize_unicode if (not isinstance(text, unicode)) else _ensure_unicode
                for k in namespace:
                    ## BAD: if isinstance(namespace[k], types.StringTypes):
                    if isinstance(namespace[k], string_types):
                        namespace[k] = transform_fn(namespace[k])
                    ## OLD: elif type(namespace[k]) in [list, dict]:
                    elif isinstance(namespace[k], (list, dict)):
                        namespace[k] = transform_fn(to_string(namespace[k]))
            # Do the actual conversion
            result = text.format(**namespace)
    except (AttributeError, KeyError, UnicodeDecodeError, ValueError):
        debug_print("Exception during format(%s,...): %s" % (text, str(sys.exc_info())), 5)
        if not ignore_exception:
            raise
    finally:
        if frame:
            del frame
    ## OLD: result = _normalize_unicode(result)
    debug_trace("txt: " + text, level=91)
    debug_trace("res: " + result, level=91)
    debug_trace("format(%s,...) => %s", text, result, level=91)
    return result


def init_logging():
    """Enable logging with INFO level by default or with DEBUG if detailed debugging"""
    debug_print("init_logging", 4)
    # TODO: use mapping from symbolic LEVEL user option (e.g., via getenv)
    level = logging.DEBUG if detailed_debugging() else logging.INFO
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=level)
    logging.debug("init_logging")
    return


def load_object(filename):
    """Load object data from FILENAME in pickle format"""
    debug_print("Loading object from %s" % filename, 3)
    object_data = None
    f = system.open_file(filename, 'rb')
    if f:
        object_data = pickle.load(f)
        f.close()
    return object_data


def store_object(filename, object_data):
    """Store OBJECT_DATA in FILENAME using pickle format"""
    debug_print("Saving object to %s" % filename, 3)
    f = system.open_file(filename, 'wb')
    if f:
        pickle.dump(object_data, f)
        f.close()
    return


def dump_stored_object(object_filename, dump_filename, level=1):
    """Traces out object stored in OBJECT_FILENAME to DUMP_FILENAME"""
    # Note: convenience function for interactive usage
    redirect_stderr(dump_filename)
    obj = load_object(object_filename)
    ## OLD: if isinstance(obj, dict) or isinstance(obj, list):
    if isinstance(obj, (dict, list)):
        trace_value(obj, level)
    else:
        trace_object(obj, level)
    restore_stderr()
    return

def create_lookup_table(filename, use_linenum=False):
    """Create lookup hash table from string keys to string values (one pair per line, tab separated), optionally with the line number serving as implicit key. Note: The keys are made lowercase, and lines with multiple tabs are ignored."""
    # TODO: rename as create_lookup_hash? (see ShelveLookup.from_hash in table_lookup.py)
    # TODO: use enumerate(f); refine exception in except
    debug_print("create_lookup_table(%s)" % filename, 4)
    lookup_hash = {} if (not use_linenum) else OrderedDict()
    f = None
    try:
        f = system.open_file(filename)
        line_num = 0
        for line in f:
            line = line.strip("\n")
            line_num += 1
            fields = line.split("\t")
            if len(fields) == 2:
                key = fields[0].lower()
                lookup_hash[key] = fields[1]
            elif (len(fields) == 1) and use_linenum:
                key = line_num
                lookup_hash[key] = fields[0]
            else:
                debug_print("Warning: Ignoring entry at line %d (%s): %s" % (line_num, filename, line), 3)
    except (IOError, ValueError):
        debug_print("Warning: Exception creating lookup table from %s: %s" % (filename, str(sys.exc_info())), 2)
    finally:
        if f:
            f.close()
    debug_print("create_lookup_table => %s" % lookup_hash, 8)
    return lookup_hash


def lookup_key(table, key, default):
    """Looks up KEY in table (ensuring lowercase)"""
    result = table.get(key.lower(), default)
    debug_format("lookup_key(_, {k}, {d}) => {r}", 6,
                 k=key, d=default, r=result)
    return result

def create_boolean_lookup_table(filename):
    """Create lookup hash table from string keys to boolean occurrence indicator. Note: The keys are made lowercase."""
    # TODO: allow for tab-delimited value to be ignored
    debug_print("create_boolean_lookup_table(%s)" % filename, 4)
    lookup_hash = {}
    f = system.open_file(filename)
    for line in f:
        key = line.strip().lower()
        lookup_hash[key] = True
    if f:
        f.close()
    debug_print("create_boolean_lookup_table => %s" % lookup_hash, 8)
    return lookup_hash


def normalize_frequencies(dictionary):
    """Normalize the frequencies in DICTIONARY (i.e., in range [0, 1])"""
    total = 0
    for k in dictionary:
        total += dictionary[k]
    if total > 0:
        for k in dictionary:
            dictionary[k] /= total
    return


def sort_frequencies(dictionary):
    """Returns the keys in DICTIONARY sorted in reverse by frequency"""
    sorted_keys = sorted(dictionary.keys(), reverse=True, 
                         key=lambda k: dictionary[k])
    debug_format("sort_frequencies({d}) => {r}", 7,
                 d=dictionary, r=sorted_keys)
    return sorted_keys


def sort_weighted_hash(weights, max_num=None, reverse=None):
    """sorts the enties in WEIGHTS hash, returns list of (key, freq) tuples.
    Note: sorted in REVERSE order by default"""
    if max_num is None:
        max_num = len(weights)
    if reverse is None:
        reverse = True
    sorted_keys = sorted(weights.keys(), reverse=reverse, 
                         key=lambda k: weights[k])
    top_values = [(k, weights[k]) for k in sorted_keys[:max_num]]
    debug_format("sort_weighted_hash(_, _) => {r}", 5, r=top_values)
    return top_values


def format_freq_hash(freqs, label, max_num=None, prec=0, indent="\t"):
    """Returns formatted string representing top entries in FREQS hash,
    preceded by LABEL, using up to MAX_NUM entries, with numbers rounded to
    PRECISION places
    Note; Result is covertedto UTF-8 if necessary (e.g., for printing)"""
    if max_num is None:
        max_num = len(freqs)
    sorted_keys = sorted(freqs.keys(), reverse=True, key=lambda k: freqs[k])
    top_values = [(indent + k + "\t" + round_num(freqs[k], prec))
                  for k in sorted_keys[:max_num]]
    hash_listing = normalize_unicode(label + '\n' + "\n".join(top_values))
    return hash_listing


def union(list1, list2):
    """"Returns set union of LIST1 and LIST2 (preserving order)"""
    # TODO: rework in terms of built-in set operations???
    result = list1[:]
    for item in list2:
        if item not in list1:
            result.append(item)
    debug_format("union({l1}, {l2}) => {r}", 7, l1=list1, l2=list2, r=result)
    return result


def intersection(list1, list2):
    """Returns set intersection of LIST1 and LIST2. Note: useful for tracing"""
    ## TODO; return system.intersection(list1, list2)
    assert(isinstance(list1, list) and isinstance(list2, list))
    result = list(set.intersection(set(list1), set(list2)))
    debug_format("intersection({l1}, {l2}) => {r}", 7, l1=list1, l2=list2, r=result)
    return result


def is_subset(list1, list2):
    """Returns whether LIST1 is subset of LIST2"""
    # EX: is_subset(['mouse', 'dog'], ['dog', 'cat', 'mouse'])
    result = set(list1).issubset(set(list2))
    debug_format("subset({l1}, {l2}) => {r}", 7, l1=list1, l2=list2, r=result)
    return result


def difference(list1, list2):
    """Returns list of items in LIST1 not in LIST2"""
    # EX: difference([5, 4, 3, 2, 1], [4, 2]) => [1, 3, 5]
    assert(isinstance(list1, list) and isinstance(list2, list))
    diff = list(set(list1).difference(set(list2)))
    debug_format("difference({l1}, {l2}) => {r}", 7, l1=list1, l2=list2, r=diff)
    return diff


def remove_all(list1, list2, ignore_case=False):
    """Removes all members of LIST2 from LIST1, similar to
    difference except that the order is preserverd"""
    # EX: remove_all([5, 4, 3, 2, 1], [4, 2, 0]) => [5, 3, 1]
    if ignore_case:
        list1 = [to_string(v).lower() for v in list1]
        list2 = [to_string(v).lower() for v in list2]
    diff = list1
    for item in list2:
        if item in diff:
            diff.remove(item)
    debug_format("remove_all({l1}, {l2}) => {r}", 7, l1=list1, l2=list2, r=diff)
    return diff


def equivalent(list1, list2):
    """Returns whether LIST1 and LIST2 have same entries (i.e., intersection is
    same as both ignoring duplicates)"""
    assert(isinstance(list1, list) and isinstance(list2, list))
    len1 = len(set(list1))
    len2 = len(set(list2))
    ok = len1 == len2 and len(intersection(list1, list2)) == len1
    debug_format("equivalent({list1}, {list1}) => {ok}", 5)
    return ok


def append_new(in_list, item):
    """Returns copy of LIST with ITEM included unless already in it"""
    result = in_list[:]
    if item not in result:
        result += [item]
    debug_print("append_new(%s, %s) => %s" % (in_list, item, result), 5)
    return result


def extract_list(text):
    """Extract list of values from TEXT, separated by spaces or commas"""
    items = text.replace(",", " ").split()
    debug_format("extract_list({text}) = {items}", 6)
    return items


def is_subsumed(term, other_terms):
    """Whether phrasal TERM is subsumed by a larger phrase in OTHER_TERMS, restricted to word boundaries"""
    # EX: is_subsumed("dog", ["dog house", "catnip"]) => True
    # EX: is_subsumed("cat", ["dog house", "catnip"]) => False
    for other_term in other_terms:
        if (term != other_term) and (" " + term + " ") in (" " + other_term + " "):
            return True
    return False


PRECISION = getenv_int("ROUNDING_PRECISION", 3,
                       "Precision for rounding numbers via round_num, etc.")
#
def round_num(num, precision=None, zero_fill=True):
    """Rounds NUM to PRECISION places (normally 3); returned as a string, optionally zero-filled on the right
    Note: Precision can be overidden wth ROUNDING_PRECISION environment variable"""
    # Note: the g conversion specifier is not used because that can yield exponent notation
    # EX: round_num(15000, 3) => "15000.000"
    # EX: round_num(15000, 3, False) => "15000"
    if precision is None:
        precision = PRECISION
    format_string = "%%.0%df" % precision
    rounded_num = (format_string % num)
    if not zero_fill:
        # Remove trailing zeros and/or trailing decimal place
        while (re.search(r"(\.[0-9]*)0$", rounded_num)):
            old_rounded_num = rounded_num
            rounded_num = re.sub(r"(\.[0-9]*)0+$", "\\1", rounded_num)
            assert(rounded_num != old_rounded_num)
        rounded_num = rounded_num.rstrip(".")
    debug_format("round_num({n}, [prec={p}, zero_fill={z}]]) => {r}", 8,
                 n=num, p=precision, z=zero_fill, r=rounded_num)
    return rounded_num


def round_nums(numbers, precision=None, zero_fill=True):
    """Rounds each of the NUMBERS to PRECISION places (normally 3): returned as list of strings, optionally zero-filled on the right"""
    # EX: round_nums([0.333333, 0.666666, 0.99999]) = ['0.333', '0.667', '1.000']
    return [round_num(v, precision, zero_fill) for v in numbers]


def round(number_or_list, precision=None):    # pylint: disable=redefined-builtin
    """Llke round_num(s) but returning float(s)"""
    # EX: round(15000) => 15000.0
    # EX: round([0.333333, 0.666666, 0.99999]) => [0.333, 0.667, 1.0]
    # Note: This is different from built-in round in that list is allowed and
    # for use of rounding-precision environment option (via round_num).
    rounded_result = 0
    ## OLD: if isinstance(number_or_list, float) or isinstance(number_or_list, int):
    if isinstance(number_or_list, (float, int)):
        rounded_result = float(round_num(number_or_list, precision))
    else:
        rounded_result = [float(s) for s in round_nums(number_or_list, precision)]
    return rounded_result


def normalize(num_list):
    """Returns NUM_LIST normalized to [0, 1] range"""
    # EX: normalize([1, 2, 3]) => [0.0, 0.5, 1.0]
    normalized = num_list
    lower = min(num_list)
    diff = float(max(num_list) - lower)
    if diff > 0:
        normalized = [(x - lower)/diff for x in num_list]
    else:
        normalized = [1.0] * len(num_list)
    debug_format("normalize({l}) => {r}", 7,
                 l=round(num_list), r=round(normalized))
    return normalized


def is_numeric(text):
    """Indicates whether TEXT represents a number (integer or float)"""
    ## TODO: rename as is_number to avoid confusion with English numeric a la numeral (e.g., "two')
    # EX: is_numeric("123") => True
    # EX: is_numeric("one") => False
    numeric = False
    try:
        _value = float(text)
        numeric = True
    except ValueError:
        _value = None
    debug_format("is_numeric({t}) => {n}; value={v}", 8, t=text, n=numeric, v=_value)
    return numeric


## OLD:
## 
## def safe_int(numeric, default_value=0, base=10):
##     """Returns NUMERIC interpreted as a int, using DEFAULT_VALUE (0) if not
##     numeric and interpretting value as optional BASE (10)"""
##     try:
##         result = int(numeric, base)
##     except (TypeError, ValueError):
##         debug_format("Warning: Exception converting integer numeric ({num}): {exc}", 4,
##                      num=numeric, exc=str(sys.exc_info()))
##         result = default_value
##     debug_format("safe_int({n}, [{df}, {b}]) => {r})", 7,
##                  n=numeric, df=default_value, b=base, r=result)
##     return result
## 
## 
## def safe_float(numeric, default_value=0.0):
##     """Returns NUMERIC interpreted as a float, using DEFAULT_VALUE (0.0) if not numeric"""
##     try:
##         result = float(numeric)
##     except ValueError:
##         debug_format("Warning: Exception converting float numeric ({num}): {exc}", 4,
##                      num=numeric, exc=str(sys.exc_info()))
##         result = default_value
##     debug_format("safe_int({n}, {df}) => {r})", 7,
##                  n=numeric, df=default_value, r=result)
##     return result
##
## def reference_variables(*args):
##     """Dummy function used for referencing variables"""
##     debug_print("reference_variables%s" % str(args), 9)
##     return
##

safe_int = system.to_int

safe_float = system.to_float

reference_variables = debug.reference_var

#------------------------------------------------------------------------
# Memomization support (i.e., functiona result caching), based on 
#     See http://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-. [world!]
# This is implemented transparently via Python decorators. See
#     http://stackoverflow.com/questions/739654/understanding-python-decorators
#
# usage example:
#
#    @memodict
#    def fubar(word):
#        """Fouled up"""
#        result = ...
#        return result
#

def memodict(f):
    """Memoization decorator for a function taking a single argument
    Warning: deprecated; uses version in system.py instead
    """
    return system.memodict(f)
    ## OLD
    ## """Memoization decorator for a function taking a single argument"""
    ## class _memodict(dict):
    ##     """Internal class for implementing memoization"""
    ##     #
    ##     def __missing__(self, key):
    ##         """Invokes function to produce value if arg not in hash"""
    ##         ret = self[key] = f(key)
    ##         return ret
    ## #
    ## return _memodict().__getitem__

#------------------------------------------------------------------------

def dummy_main():
    """Dummy entry point for script, showing environment variable usage"""
    if detailed_debugging():
        # Create dummy environment opton with description
        _FUBAR = getenv_boolean("FUBAR", True, "Dummy option")
    print("Environment options:")
    desc = formatted_environment_option_descriptions(include_all=True)
    print("\t" + (desc if desc.strip() else "n/a"))
    return


# Do debug-only processing (n.b., for when PYTHONOPTIMIZE not set)
# Note: wrapped in function to keep things clean

if __debug__:

    def debug_init():
        """Debug-only initialization"""
        ## OLD: tpo_common_start = time.time()

        # Override debug level to DEBUG_LEVEL environment variable unless running in optimized mode (i.e., __debug__ is False)
        ## OLD:
        ## env_debug_level = os.getenv("DEBUG_LEVEL")
        ## global debug_level
        ## debug_level = int(env_debug_level) if env_debug_level else ERROR
    
        # Show trace level in effect
        debug_format("starting tpo_common.py at {ts}: "
                     + "debug_level={lvl}; args={args}", level=DETAILED,
                     ts=debug_timestamp(), lvl=debugging_level(), args=sys.argv)

        ## OLD
        ## # Register DEBUG_LEVEL for sake of new users
        ## test_debug_level = getenv_integer("DEBUG_LEVEL", debugging_level(), 
        ##                                   "Debugging level for script tracing")
        ## assert(debugging_level() == test_debug_level)

        # Register function to display ending timestamp
        ## OLD:
        ## def _display_ending_info():
        ##     """Display ending time information"""
        ##     # note: does nothing if stderr closed (e.g., during profiling)
        ##     # TODO: track down problem with pytest
        ##     if sys.stderr.closed:       # pylint: disable=using-constant-test
        ##         return
        ##     trace_object(sys.stderr, level=7)
        ##     tpo_common_end = time.time()
        ##     debug_format("ending tpo_common.py at {time}; elapsed={diff}s", 
        ##                  level=DETAILED, time=debug_timestamp(), 
        ##                  diff=round_num(tpo_common_end - tpo_common_start))
        ## #
        ## if not getenv_boolean("OMIT_ATEXIT", False):
        ##     atexit.register(_display_ending_info)

        # Override globals from environment
        # Note: done after start-up tracing initialization for getenv tracing
        ## OLD:
        ## global output_timestamps
        ## output_timestamps = getenv_boolean("OUTPUT_DEBUG_TIMESTAMPS", 
        ##                                    output_timestamps)
        ## global use_logging
        ## use_logging = getenv_boolean("USE_LOGGING", use_logging)
        ## enable_logging = getenv_boolean("ENABLE_LOGGING", use_logging)
        ## if enable_logging:
        ##     init_logging()
        global skip_format_warning
        skip_format_warning = getenv_boolean("SKIP_FORMAT_WARNING", False)
    
        # Warn about old version of format
        if USE_SIMPLE_FORMAT:
            debug_print("Warning: Using simple format for debug_format", 
                        level=WARNING)

        return
    
    # Do the initialization
    debug_init()
    debug.assertion(debug_level == debug.get_level())

# Warn if invoked standalone
#
if __name__ == '__main__':
    print_stderr("Warning: %s is not intended to be run standalone" % __file__)
    dummy_main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Test(s) for ../llm_desktop_search.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_llm_desktop_search.py
#
# TODO3:
# - Try to minize usage of run_script to just one or two tests:
#   it is an older style of testing. It is better to use DesktopSearch
#   class directly. More details follow in the warning.
#
# Warning:
# - The use of run_script as in test_01_data_file is an older style of testing.
#   It is better to directly invoke a helper class in the script that is independent
#   of the Script class based on Main, which is mainly for argument parsing.
#   (For an example of this, see python_ast.py and tests/tests_python_ast.py.)
# - Moreover, debugging tests with run_script is complicated because a separate
#   process is involved (e.g., with separate environment variables).
# - See discussion of SUB_DEBUG_LEVEL in unittest_wrapper.py for more info.
#

"""Tests for llm_desktop_search module"""

# Standard modules
import atexit

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper, RUN_SLOW_TESTS, invoke_tests
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import gpu_utils
from mezcla.main import KEEP_TEMP_FILES
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        module instance (e.g,, <module 'mezcla.main' from '/home/testuser/Mezcla/mezcla/main.py'>
#    TestIt.script_module:              dotted module path (e.g., "mezcla.main")
try:
    import mezcla.llm_desktop_search as THE_MODULE
    get_last_modified_date = THE_MODULE.get_last_modified_date
except:
    THE_MODULE = None
    get_last_modified_date = None
    debug.trace_exception(3, "llm_desktop_search import")

#------------------------------------------------------------------------


@pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    INDEX_STORE_DIR = (THE_MODULE.INDEX_STORE_DIR.lstrip().lstrip(system.path_separator())
                       if THE_MODULE else None)
    use_temp_base_dir = True            # needed for self.temp_base to be a dir
    
    # set a temp dir to test index indexing setUpClass
    # Note: index_temp_dir needs to be unique
    index_temp_dir = None
    index_parent = None

    @classmethod
    def setUpClass(cls, filename=None, module=None):
        """One-time initialization (i.e., for entire class)"""
        # note: creates default FAISS index shared by tests (unless overriden)
        debug.trace(6, f"TestIt.setUpClass(); cls={cls}")
        # note: should do parent processing first
        super().setUpClass(filename, module)
        cls.index_parent = cls.temp_base
        cls.index_temp_dir = gh.form_path(cls.index_parent, cls.INDEX_STORE_DIR)
        if not system.is_directory(cls.index_temp_dir):
            gh.full_mkdir(cls.index_temp_dir)
        if THE_MODULE.INDEX_ONLY_RECENT:
            THE_MODULE.INDEX_ONLY_RECENT = False
        debug.trace_object(5, cls, label=f"{cls.__class__.__name__} instance")
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="Ignoring slow test")
    def test_01_index_dir(self):
        """Tests run_script to index directory"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_index_dir(); self={self}")
        if not KEEP_TEMP_FILES:
            atexit.register(gh.delete_directory, self.index_temp_dir)
            
        if not system.is_directory(self.index_parent):
           debug.assertion(False)
           gh.full_mkdir(self.index_parent)
        
        # test if indexing works with with no existing db
        file_dir = gh.real_path(gh.dirname(__file__))
        repo_base_dir = gh.form_path(file_dir, "..", "..")
        
        init_output = self.run_script(options=f"--index {repo_base_dir}",
                                      env_options=f"INDEX_STORE_DIR={self.index_temp_dir}")
        self.do_assert(my_re.search(r"(\d\d+) chunks indexed", init_output))
        num_initial_chunks = int(my_re.group(1))
        index_files = system.read_directory(self.index_temp_dir)
        
        # assert INDEX_STORE_DIR is not empty
        self.do_assert(index_files != [])
        
        # save modified date for comparing later 
        prev_date = get_last_modified_date(system.get_directory_filenames(self.index_temp_dir, just_regular_files=True))
        
        # test that indexing with an already existing DB works
        resource_dir = gh.form_path(file_dir, "resources")
        revised_output = self.run_script(options=f"--index {resource_dir}",
                                         env_options=f"INDEX_STORE_DIR={self.index_temp_dir}")
        self.do_assert(my_re.search(r"(\d\d+) chunks indexed", init_output))
        num_final_chunks = int(my_re.group(1))
        self.do_assert(num_final_chunks > num_initial_chunks)
       
        # get modification time and check if it changed
        new_date = get_last_modified_date(system.get_directory_filenames(self.index_temp_dir, just_regular_files=True))
        self.do_assert(new_date > prev_date)
        

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="Ignoring slow test")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_search_docs(self):
        """Test for something_else: TODO..."""
        debug.trace(4, f"TestIt.test_02_search_docs(); self={self}")
        desktop = THE_MODULE.DesktopSearch(self.index_temp_dir)
        desktop.search_to_answer('What license is used?')
        output = self.get_stdout()
        
        self.do_assert(my_re.search(r"GNU", output.strip()))
        

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(gpu_utils.TORCH_DEVICE != "cuda", reason="Ignoring non-CUDA device")
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="Ignoring slow test")
    def test_03_gpu_usage(self):
        """Test for GPU libs being used"""
        # Note: verifies process using GPU via nvida-smi
        #  |  GPU   GI   CI        PID   Type   Process name            GPU Memory |
        #  ...
        #  |    0   N/A  N/A   1111609      C   python                      366MiB |
        debug.trace(4, f"TestIt.test_03_gpu_usage(); self={self}")
        ds = THE_MODULE.DesktopSearch(self.index_temp_dir)
        ds.show_similar("license")
        trace_level = max(1, debug.get_level())
        gpu_utils.trace_gpu_usage(level=trace_level)
        stdout, stderr = self.get_stdout_stderr()
        self.do_assert("GNU" in stdout) 
        pid = system.get_process_id()
        self.do_assert(my_re.search(fr"\b{pid}\b.*python", stderr))
        return
    

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="Ignoring slow test")
    def test_04_show_similar(self):
        """Test run_script to show similar document to QUERY"""
        debug.trace(4, f"test_04_show_similar(): self={self}")
        desktop = THE_MODULE.DesktopSearch(index_store_dir=self.index_temp_dir)
        
        desktop.show_similar(query="LICENSE", num=1)
        output = self.get_stdout()
        if not KEEP_TEMP_FILES:
            gh.delete_directory(self.index_temp_dir)
        self.do_assert("Lesser General Public License" in output)
        

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="Ignoring slow test")
    def test_05_index_via_API(self):
        """Run indexing via class-based API"""
        debug.trace(4, f"test_05_index_via_API(): self={self}")
        
        # Index the files
        # note: index is specific to this test case
        temp_index_dir = gh.form_path(self.temp_base, "index")
        test_dir = gh.dirname(__file__)
        doc_dir = gh.resolve_path("resources", base_dir=test_dir)
        ds = THE_MODULE.DesktopSearch(index_store_dir=temp_index_dir)
        ds.index_dir(doc_dir)

        # Make sure most of expected content got indexed included (e.g., 75%)
        #
        # Note: document contents check as follows
        # In [67]: docid = db.index_to_docstore_id[0];  db.docstore.search(docid).page_content
        # Out[67]: 'Tío Tomás\t\t\t\tUncle Tom\n\n¡Buenos días!\t\t\t\tGood morning\n\nçãêâôöèäàÃëÇÂîòïÔìðÊÅåùÀŠý\t\tcaeaooeaaAeCAioiOioEAauASy'
        #
        expected_text = ["Tío Tomás", "Library", "validation", "Iris", "Argentina"]
        num_total = len(expected_text)
        num_found = 0
        for text in expected_text:
           for docid in ds.db.index_to_docstore_id.values():
              if text in ds.db.docstore.search(docid).page_content:
                 num_found += 1
                 break
        pct_75 = (3 * num_total // 4)
        debug.trace_expr(5, num_found, num_total, pct_75)
        assert(num_found >= pct_75)

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)
